# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the collection models."""

import datetime as dt
from dataclasses import dataclass
from typing import Any, ClassVar, NamedTuple, override
from urllib.parse import quote

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import connection, transaction
from django.db.utils import IntegrityError
from django.test import RequestFactory
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebusinePromise,
)
from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    CollectionRelation,
    Scope,
    User,
    WorkRequest,
    Workspace,
)
from debusine.db.models.collections import CollectionRoleBase, CollectionRoles
from debusine.db.models.tests.utils import RolesTestCase
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.server.collections import DebianSuiteManager
from debusine.tasks.models import BackendType
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    override_permission,
)


class CollectionRolesTests(RolesTestCase[CollectionRoles]):
    """Tests for the CollectionRoles class."""

    scenario = scenarios.DefaultContext()
    roles_class = CollectionRoles

    def test_invariants(self) -> None:
        self.assertRolesInvariants()

    def test_from_iterable(self) -> None:
        for value, expected in (
            ([], []),
            (["owner"], [Collection.Roles.OWNER]),
            (["viewer"], [Collection.Roles.VIEWER]),
            ([Collection.Roles.OWNER], [Collection.Roles.OWNER]),
            ([Collection.Roles.VIEWER], [Collection.Roles.VIEWER]),
            (["owner", "viewer"], [Collection.Roles.OWNER]),
            (
                [
                    "owner",
                    "owner",
                    Collection.Roles.OWNER,
                    Collection.Roles.VIEWER,
                ],
                [Collection.Roles.OWNER],
            ),
        ):
            with self.subTest(value=repr(value)):
                self.assertEqual(
                    Collection.Roles.from_iterable(value),
                    frozenset(expected),
                )

    def test_choices(self) -> None:
        self.assertChoices()

    def test_q_non_authenticated(self) -> None:
        """Test generating Q objects with anonymous users."""
        standard_collections = list(
            Collection.objects.filter(workspace=self.scenario.workspace)
        )
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        for role, expected in (
            (Collection.Roles.OWNER, False),
            (Collection.Roles.VIEWER, True),
        ):
            with self.subTest(role=repr(role)):
                expected_qs = (
                    [collection, *standard_collections] if expected else []
                )
                self.assertQuerySetEqual(
                    Collection.objects.filter(
                        workspace=self.scenario.workspace
                    ).filter(role.q(PermissionContext.for_ui_anonymous())),
                    expected_qs,
                    ordered=False,
                )

    def test_values(self) -> None:
        role = Collection.Roles.OWNER
        self.assertEqual(role.label, "Owner")
        self.assertEqual(
            role.implied_by_scope_roles, frozenset([Scope.Roles.OWNER])
        )
        self.assertEqual(
            role.implied_by_workspace_roles, frozenset([Workspace.Roles.OWNER])
        )
        self.assertEqual(role.implied_by_collection_roles, frozenset([role]))
        self.assertTrue(role.implies(Collection.Roles.VIEWER))

        role = Collection.Roles.VIEWER
        self.assertEqual(role.label, "Viewer")
        self.assertEqual(
            role.implied_by_scope_roles,
            frozenset([Scope.Roles.OWNER, Scope.Roles.OBSERVER]),
        )
        self.assertEqual(
            role.implied_by_workspace_roles,
            frozenset(
                [
                    Workspace.Roles.OWNER,
                    Workspace.Roles.CONTRIBUTOR,
                    Workspace.Roles.VIEWER,
                ]
            ),
        )
        self.assertEqual(
            role.implied_by_collection_roles,
            frozenset([role, Collection.Roles.OWNER]),
        )
        self.assertFalse(role.implies(Collection.Roles.OWNER))

    def test_init_invalid_implication(self) -> None:
        r = CollectionRoleBase("test", implied_by=[None])
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"Collection roles do not support implications by None",
        ):
            r._setup()


class CollectionManagerTests(TestCase):
    """Tests for CollectionManager class."""

    scenario = scenarios.DefaultContext()

    standard_collections: ClassVar[list[Collection]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.standard_collections = list(
            Collection.objects.filter(workspace=cls.scenario.workspace)
        )

    def test_create_singleton(self) -> None:
        """Create a singleton collection."""
        workspace = self.playground.create_workspace(name="test")
        collection, _ = Collection.objects.get_or_create_singleton(
            category=CollectionCategory.PACKAGE_BUILD_LOGS,
            workspace=workspace,
        )
        self.assertEqual(collection.name, "_")
        self.assertEqual(
            collection.category, CollectionCategory.PACKAGE_BUILD_LOGS
        )
        self.assertEqual(collection.workspace, workspace)
        self.assertEqual(collection.data, {})

    def test_create_singleton_bad_category(self) -> None:
        """Singleton collections may only be of certain categories."""
        with self.assertRaisesRegex(
            ValueError,
            f"'{CollectionCategory.TEST}' is not a singleton collection "
            f"category",
        ):
            Collection.objects.get_or_create_singleton(
                category=CollectionCategory.TEST,
                workspace=self.scenario.workspace,
            )

    def test_in_current_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter."""
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )

        with context.local():
            self.scenario.set_current()
            self.assertQuerySetEqual(
                Collection.objects.in_current_scope(),
                [collection, *self.standard_collections],
                ordered=False,
            )

        scope1 = self.playground.get_or_create_scope(name="Scope1")
        with context.local():
            context.set_permission_context_scoped(
                scope1, PermissionContext.for_ui_anonymous()
            )
            self.assertQuerySetEqual(
                Collection.objects.in_current_scope(),
                [],
            )

    def test_in_current_scope_no_context_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter without scope set."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "scope is not set"
        ):
            Collection.objects.in_current_scope()

    def test_in_current_workspace(self) -> None:
        """Test the in_current_workspace() QuerySet filter."""
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )

        with context.local():
            self.scenario.workspace.set_current()
            self.assertQuerySetEqual(
                Collection.objects.in_current_workspace(),
                [collection, *self.standard_collections],
                ordered=False,
            )

        workspace1 = self.playground.create_workspace(name="other", public=True)
        with context.local():
            workspace1.set_current()
            self.assertQuerySetEqual(
                Collection.objects.in_current_workspace(),
                [],
            )

    def test_in_current_workspace_no_context_workspace(self) -> None:
        """Test in_current_workspace() without workspace set."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "workspace is not set"
        ):
            Collection.objects.in_current_workspace()


class CollectionTests(TestCase):
    """Tests for Collection class."""

    scenario = scenarios.DefaultContext()

    standard_collections: ClassVar[list[Collection]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.standard_collections = list(
            Collection.objects.filter(workspace=cls.scenario.workspace)
        )

    def assertHasRoles(
        self,
        roles: list[Collection.Roles],
        collection: Collection,
        *,
        pc: PermissionContext | None = None,
    ) -> None:
        """Ensure the user has exactly the given roles on the collection."""
        pc = pc or self.scenario.context_ui()
        found = [
            role for role in Collection.Roles if collection.has_role(pc, role)
        ]
        self.assertEqual(
            ", ".join(str(r) for r in found), ", ".join(str(r) for r in roles)
        )

    def test_str(self) -> None:
        """Test stringification."""
        name = "Environments"
        category = CollectionCategory.ENVIRONMENTS
        collection = Collection.objects.create(
            name=name, category=category, workspace=self.scenario.workspace
        )
        self.assertEqual(str(collection), f"{name}@{category}")

        singleton = Collection.objects.get(
            category=CollectionCategory.PACKAGE_BUILD_LOGS,
            workspace=self.scenario.workspace,
        )
        self.assertEqual(
            str(singleton), f"_@{CollectionCategory.PACKAGE_BUILD_LOGS}"
        )

    def test_ui(self) -> None:
        from debusine.web.views.ui.collections import CollectionUI

        collection = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        request = RequestFactory().get("/")
        ui = collection.ui(request)
        self.assertIsInstance(ui, CollectionUI)

    def test_grant_role(self) -> None:
        collection = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        token = self.playground.create_external_user_token(username="test")
        grant = collection.grant_role(
            token, Collection.Roles.VIEWER, created_by=self.scenario.user
        )

        self.assertEqual(grant.token, token)
        self.assertEqual(grant.collection, collection)
        self.assertEqual(grant.role, Collection.Roles.VIEWER)
        self.assertLessEqual(grant.created_at, timezone.now())
        self.assertEqual(grant.created_by, self.scenario.user)

    def test_has_role(self) -> None:
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        OWNER = Collection.Roles.OWNER
        VIEWER = Collection.Roles.VIEWER
        for scope_roles, workspace_roles, collection_roles, roles, queries in (
            ([], [], [], [], 2),
            ([], [], [OWNER], [OWNER, VIEWER], 2),
            ([], [], [VIEWER], [VIEWER], 2),
            ([Scope.Roles.OWNER], [], [], [OWNER, VIEWER], 2),
            ([Scope.Roles.OBSERVER], [], [], [VIEWER], 2),
            ([], [Workspace.Roles.OWNER], [], [OWNER, VIEWER], 2),
            ([], [Workspace.Roles.CONTRIBUTOR], [], [VIEWER], 2),
            ([], [Workspace.Roles.VIEWER], [], [VIEWER], 2),
        ):
            with (
                self.subTest(
                    scope_roles=repr(scope_roles),
                    workspace_roles=repr(workspace_roles),
                    collection_roles=repr(collection_roles),
                ),
                self.scenario.assign_role(self.scenario.scope, *scope_roles),
                self.scenario.assign_role(
                    self.scenario.workspace, *workspace_roles
                ),
                self.scenario.assign_role(collection, *collection_roles),
            ):
                with self.subTest(type="query"), self.assertNumQueries(queries):
                    self.assertHasRoles(roles, collection)
                if scope_roles or workspace_roles:
                    with self.subTest(type="context"), context.local():
                        self.scenario.set_current()
                        with self.assertNumQueries(
                            0
                            if Scope.Roles.OWNER in scope_roles
                            or Workspace.Roles.OWNER in workspace_roles
                            else 1
                        ):
                            self.assertHasRoles(roles, collection)

    def test_get_absolute_url(self) -> None:
        """Test get_absolute_url results."""
        name = "Environments"
        category = CollectionCategory.ENVIRONMENTS
        collection = Collection.objects.create(
            name=name, category=category, workspace=self.scenario.workspace
        )

        self.assertEqual(
            collection.get_absolute_url(),
            f"/{self.scenario.scope.name}/{self.scenario.workspace.name}"
            f"/collection/{CollectionCategory.ENVIRONMENTS}/{name}/",
        )

    def test_get_absolute_url_different_scope(self) -> None:
        """get_absolute_url works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        name = "Environments"
        category = CollectionCategory.ENVIRONMENTS
        collection = Collection.objects.create(
            name=name, category=category, workspace=other_workspace
        )

        self.assertEqual(
            collection.get_absolute_url(),
            f"/{other_scope.name}/{other_workspace.name}"
            f"/collection/{CollectionCategory.ENVIRONMENTS}/{name}/",
        )

    def test_get_absolute_url_search(self) -> None:
        """Test get_absolute_url_search results."""
        name = "Environments"
        category = CollectionCategory.ENVIRONMENTS
        collection = Collection.objects.create(
            name=name, category=category, workspace=self.scenario.workspace
        )

        self.assertEqual(
            collection.get_absolute_url_search(),
            f"/{self.scenario.scope.name}/{self.scenario.workspace.name}"
            f"/collection/{CollectionCategory.ENVIRONMENTS}/{name}/search/",
        )

    def test_get_absolute_url_search_different_scope(self) -> None:
        """get_absolute_url_search works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        name = "Environments"
        category = CollectionCategory.ENVIRONMENTS
        collection = Collection.objects.create(
            name=name, category=category, workspace=other_workspace
        )

        self.assertEqual(
            collection.get_absolute_url_search(),
            f"/{other_scope.name}/{other_workspace.name}"
            f"/collection/{CollectionCategory.ENVIRONMENTS}/{name}/search/",
        )

    def test_constraint_name_category_workspace(self) -> None:
        """Raise integrity error: name, category, workspace must be unique."""
        name = "Environments"
        category = CollectionCategory.ENVIRONMENTS

        Collection.objects.create(
            name=name, category=category, workspace=self.scenario.workspace
        )

        with (
            transaction.atomic(),
            self.assertRaisesRegex(
                IntegrityError, "db_collection_unique_name_category_workspace"
            ),
        ):
            Collection.objects.create(
                name=name, category=category, workspace=self.scenario.workspace
            )

        other_workspace = self.playground.create_workspace(name="other")
        Collection.objects.create(
            name=name, category=category, workspace=other_workspace
        )

    def test_name_not_empty(self) -> None:
        """Cannot create Collection with an empty name."""
        with self.assertRaisesRegex(
            IntegrityError, "db_collection_name_not_empty"
        ):
            Collection.objects.create(
                name="",
                category=CollectionCategory.ENVIRONMENTS,
                workspace=self.scenario.workspace,
            )

    def test_name_validation(self) -> None:
        Collection(
            name="foo",
            category=CollectionCategory.TEST,
            workspace=self.scenario.workspace,
        ).full_clean()
        Collection(
            name="foo_",
            category=CollectionCategory.TEST,
            workspace=self.scenario.workspace,
        ).full_clean()
        with self.assertRaises(ValidationError) as exc:
            Collection(
                name="_foo",
                category=CollectionCategory.TEST,
                workspace=self.scenario.workspace,
            ).full_clean()
        self.assertEqual(
            exc.exception.message_dict,
            {'name': ["'_foo' is not a valid collection name"]},
        )
        Collection.objects.get_or_create_singleton(
            category=CollectionCategory.ARCHIVE,
            workspace=self.scenario.workspace,
        )[0].full_clean()

    def test_category_not_empty(self) -> None:
        """Cannot create Collection with an empty category."""
        with self.assertRaisesRegex(
            IntegrityError, "db_collection_category_not_empty"
        ):
            Collection.objects.create(
                name="name", category="", workspace=self.scenario.workspace
            )

    def test_retains_artifacts(self) -> None:
        """By default, a new Collection stops its artifacts expiring."""
        collection = Collection.objects.create(
            name="Environments",
            category=CollectionCategory.ENVIRONMENTS,
            workspace=self.scenario.workspace,
        )
        self.assertEqual(
            collection.retains_artifacts, Collection.RetainsArtifacts.ALWAYS
        )

    def test_manager(self) -> None:
        """Test getting specialized manager."""
        collection = Collection.objects.create(
            name="Testing",
            category=CollectionCategory.SUITE,
            workspace=self.scenario.workspace,
        )
        self.assertEqual(type(collection.manager), DebianSuiteManager)

    def test_child_items_artifacts_collections(self) -> None:
        """Child items returns expected CollectionItems, artifacts, etc."""
        collection_parent_1 = Collection.objects.create(
            name="collection-1",
            category="collection-1",
            workspace=self.scenario.workspace,
        )
        collection_parent_2 = Collection.objects.create(
            name="collection-2",
            category="collection-2",
            workspace=self.scenario.workspace,
        )

        collection = Collection.objects.create(
            name="collection-3",
            category="collection-3",
            workspace=self.scenario.workspace,
        )
        collection_item_collection_1 = CollectionItem.objects.create(
            name="test",
            category="test",
            parent_collection=collection_parent_1,
            child_type=CollectionItem.Types.COLLECTION,
            collection=collection,
            created_by_user=self.scenario.user,
        )

        artifact, _ = self.playground.create_artifact()
        artifact_item_collection_1 = CollectionItem.objects.create(
            name="test-2",
            category="test-2",
            parent_collection=collection_parent_1,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact,
            created_by_user=self.scenario.user,
        )

        # Used to see that is not returned when querying
        # collection_parent_1.child_items.all()
        CollectionItem.objects.create(
            name="test",
            category="test",
            parent_collection=collection_parent_2,
            child_type=CollectionItem.Types.COLLECTION,
            collection=Collection.objects.create(
                name="collection-4",
                category="collection-3",
                workspace=self.scenario.workspace,
            ),
            created_by_user=self.scenario.user,
        )

        # Collection.child_items returns all CollectionItems
        self.assertQuerySetEqual(
            collection_parent_1.child_items.all(),
            [collection_item_collection_1, artifact_item_collection_1],
            ordered=False,
        )

        # Collection.child_artifacts returns only the artifacts
        self.assertQuerySetEqual(
            collection_parent_1.child_artifacts.all(), [artifact]
        )

        # Collection.child_collections returns only the collections
        self.assertQuerySetEqual(
            collection_parent_1.child_collections.all(), [collection]
        )

        # Given the artifact: there is a reverse relationship
        # with the collections that belongs to
        self.assertQuerySetEqual(
            artifact.parent_collections.all(),
            [collection_parent_1],
        )

        # Given a collection: there is a reverse relationship
        # with the collections that belongs to
        self.assertQuerySetEqual(
            collection.parent_collections.all(),
            [collection_parent_1],
        )

        # Given a artifact: there is a reverse relationship
        # to the CollectionItems that belongs to
        self.assertQuerySetEqual(
            artifact.collection_items.all(),
            [artifact_item_collection_1],
        )

        # Given a collection: there is a reverse relationship
        # to the CollectionItems that belongs to
        self.assertQuerySetEqual(
            collection.collection_items.all(),
            [collection_item_collection_1],
        )

    def test_can_display_workspace_public(self) -> None:
        """Test the can_display predicate on public workspaces."""
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        self.assertPermission(
            "can_display",
            users=(AnonymousUser(), self.scenario.user),
            allowed=[collection, *self.standard_collections],
        )

    def test_can_display_workspace_private(self) -> None:
        """Test the can_display predicate on private workspaces."""
        user = self.scenario.user
        workspace = self.scenario.workspace
        standard = self.standard_collections
        workspace.public = False
        workspace.save()
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        for u, workspace_roles, collection_roles, allowed, denied in (
            (AnonymousUser(), [], [], [], [collection, *standard]),
            (user, [], [], [], [collection, *standard]),
            (user, [], [Collection.Roles.VIEWER], [collection], standard),
            (user, [], [Collection.Roles.OWNER], [collection], standard),
            (user, [Workspace.Roles.VIEWER], [], [collection, *standard], []),
            (
                user,
                [Workspace.Roles.CONTRIBUTOR],
                [],
                [collection, *standard],
                [],
            ),
            (user, [Workspace.Roles.OWNER], [], [collection, *standard], []),
        ):
            with (
                self.subTest(
                    user=repr(u),
                    workspace_roles=repr(workspace_roles),
                    collection_roles=repr(collection_roles),
                ),
                self.scenario.assign_role(workspace, *workspace_roles),
                self.scenario.assign_role(collection, *collection_roles),
            ):
                self.assertPermission(
                    "can_display", users=u, allowed=allowed, denied=denied
                )

    def test_can_display_external_user_token(self) -> None:
        """An external user token can display a particular collection."""
        public = self.playground.get_default_workspace()
        private = self.playground.create_workspace(name="private", public=False)
        collections = [
            self.playground.create_collection(
                name, CollectionCategory.TEST, workspace=private
            )
            for name in ("test", "test2")
        ]
        tokens = [
            self.playground.create_external_user_token(username=username)
            for username in ("test", "test2")
        ]
        collections[0].grant_role(
            tokens[0], Collection.Roles.VIEWER, created_by=self.scenario.user
        )
        collections[1].grant_role(
            tokens[1], Collection.Roles.VIEWER, created_by=self.scenario.user
        )
        for token, collection in zip(tokens, collections):
            with self.subTest(username=token.external_username):
                self.assertPermission(
                    "can_display",
                    users=AnonymousUser(),
                    denied=(public, private),
                    token=token,
                )
                self.assertPermission(
                    "can_display",
                    users=AnonymousUser(),
                    allowed=collection,
                    denied=tuple(Collection.objects.exclude(pk=collection.pk)),
                    token=token,
                )

    def test_can_display_external_user_token_archive(self) -> None:
        """Granting access to an archive grants access to all its suites."""
        public = self.playground.get_default_workspace()
        private = [
            self.playground.create_workspace(name=name, public=False)
            for name in ("private-1", "private-2")
        ]
        archives = [
            self.playground.create_singleton_collection(
                CollectionCategory.ARCHIVE, workspace=workspace
            )
            for workspace in private
        ]
        suites: list[Collection] = []
        for archive in archives:
            suites += [
                self.playground.create_collection(
                    name, CollectionCategory.SUITE, workspace=archive.workspace
                )
                for name in ("bookworm", "trixie")
            ]
        token = self.playground.create_external_user_token(username="test")
        archives[0].grant_role(
            token, Collection.Roles.VIEWER, created_by=self.scenario.user
        )
        self.assertPermission(
            "can_display",
            users=AnonymousUser(),
            denied=(public, *private),
            token=token,
        )
        self.assertPermission(
            "can_display",
            users=AnonymousUser(),
            allowed=(archives[0], *suites[:2]),
            denied=tuple(
                Collection.objects.exclude(
                    pk__in={
                        collection.pk
                        for collection in (archives[0], *suites[:2])
                    }
                )
            ),
            token=token,
        )

    def test_can_configure_delegate_to_workspace(self) -> None:
        """Test the can_configure predicate on public workspaces."""
        user = self.scenario.user
        workspace = self.scenario.workspace
        standard = self.standard_collections
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        for u, workspace_roles, collection_roles, allowed, denied in (
            (AnonymousUser(), [], [], [], [collection, *standard]),
            (user, [], [], [], [collection, *standard]),
            (user, [], [Collection.Roles.VIEWER], [], [collection, *standard]),
            (user, [], [Collection.Roles.OWNER], [collection], standard),
            (user, [Workspace.Roles.VIEWER], [], [], [collection, *standard]),
            (
                user,
                [Workspace.Roles.CONTRIBUTOR],
                [],
                [],
                [collection, *standard],
            ),
            (user, [Workspace.Roles.OWNER], [], [collection, *standard], []),
        ):
            with (
                self.subTest(
                    user=repr(u),
                    workspace_roles=repr(workspace_roles),
                    collection_roles=repr(collection_roles),
                ),
                self.scenario.assign_role(workspace, *workspace_roles),
                self.scenario.assign_role(collection, *collection_roles),
            ):
                self.assertPermission(
                    "can_configure", users=u, allowed=allowed, denied=denied
                )

    def test_can_delete_delegate_to_workspace(self) -> None:
        user = self.scenario.user
        workspace = self.scenario.workspace
        standard = self.standard_collections
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        for u, workspace_roles, collection_roles, allowed, denied in (
            (AnonymousUser(), [], [], [], [collection, *standard]),
            (user, [], [], [], [collection, *standard]),
            (user, [], [Collection.Roles.VIEWER], [], [collection, *standard]),
            (user, [], [Collection.Roles.OWNER], [collection], standard),
            (user, [Workspace.Roles.VIEWER], [], [], [collection, *standard]),
            (
                user,
                [Workspace.Roles.CONTRIBUTOR],
                [],
                [],
                [collection, *standard],
            ),
            (user, [Workspace.Roles.OWNER], [], [collection, *standard], []),
        ):
            with (
                self.subTest(
                    user=repr(u),
                    workspace_roles=repr(workspace_roles),
                    collection_roles=repr(collection_roles),
                ),
                self.scenario.assign_role(workspace, *workspace_roles),
                self.scenario.assign_role(collection, *collection_roles),
            ):
                self.assertPermission(
                    "can_delete", users=u, allowed=allowed, denied=denied
                )

    def test_can_create_collection_relation_delegate_to_workspace(self) -> None:
        user = self.scenario.user
        workspace = self.scenario.workspace
        standard = self.standard_collections
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        for u, workspace_roles, collection_roles, allowed, denied in (
            (AnonymousUser(), [], [], [], [collection, *standard]),
            (user, [], [], [], [collection, *standard]),
            (user, [], [Collection.Roles.VIEWER], [], [collection, *standard]),
            (user, [], [Collection.Roles.OWNER], [collection], standard),
            (user, [Workspace.Roles.VIEWER], [], [], [collection, *standard]),
            (
                user,
                [Workspace.Roles.CONTRIBUTOR],
                [],
                [],
                [collection, *standard],
            ),
            (user, [Workspace.Roles.OWNER], [], [collection, *standard], []),
        ):
            with (
                self.subTest(
                    user=repr(u),
                    workspace_roles=repr(workspace_roles),
                    collection_roles=repr(collection_roles),
                ),
                self.scenario.assign_role(workspace, *workspace_roles),
                self.scenario.assign_role(collection, *collection_roles),
            ):
                self.assertPermission(
                    "can_create_collection_relation",
                    users=u,
                    allowed=allowed,
                    denied=denied,
                )

    def test_can_delete_do_not_delete_workflow_internal(self) -> None:
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        internal_collection = self.playground.create_collection(
            name="test", category=CollectionCategory.WORKFLOW_INTERNAL
        )
        with self.scenario.assign_role(
            self.scenario.workspace, Workspace.Roles.OWNER
        ):
            self.assertPermission(
                "can_delete",
                users=self.scenario.user,
                allowed=[collection, *self.standard_collections],
                denied=internal_collection,
            )

    def test_delete_collection_empty(self) -> None:
        source = self.playground.create_source_artifact()
        sid = self.playground.create_collection(
            "sid", category=CollectionCategory.SUITE
        )
        item = sid.manager.add_artifact(
            source,
            user=self.scenario.user,
            variables={"component": "main", "section": "extra"},
        )
        test = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        test.delete_collection()
        self.assertFalse(Collection.objects.filter(pk=test.pk).exists())
        self.assertTrue(Collection.objects.filter(pk=sid.pk).exists())
        self.assertTrue(CollectionItem.objects.filter(pk=item.pk).exists())

    def test_delete_collection_with_items(self) -> None:
        source = self.playground.create_source_artifact()
        sid = self.playground.create_collection(
            "sid", category=CollectionCategory.SUITE
        )
        item = sid.manager.add_artifact(
            source,
            user=self.scenario.user,
            variables={"component": "main", "section": "extra"},
        )
        test = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        sid.delete_collection()
        self.assertFalse(Collection.objects.filter(pk=sid.pk).exists())
        self.assertFalse(CollectionItem.objects.filter(pk=item.pk).exists())
        self.assertTrue(Collection.objects.filter(pk=test.pk).exists())
        self.assertTrue(Artifact.objects.filter(pk=source.pk).exists())

    def test_delete_collection_with_subcollections(self) -> None:
        source = self.playground.create_source_artifact()
        debian = Collection.objects.get(
            name="_", category=CollectionCategory.ARCHIVE
        )
        item_collection = sid = self.playground.create_collection(
            "sid", category=CollectionCategory.SUITE
        )
        item_artifact = sid.manager.add_artifact(
            source,
            user=self.scenario.user,
            variables={"component": "main", "section": "extra"},
        )
        debian.manager.add_collection(sid, user=self.scenario.user)
        debian.delete_collection()
        self.assertFalse(Collection.objects.filter(pk=debian.pk).exists())
        self.assertFalse(
            CollectionItem.objects.filter(pk=item_collection.pk).exists()
        )
        self.assertTrue(Collection.objects.filter(pk=sid.pk).exists())
        self.assertTrue(
            CollectionItem.objects.filter(pk=item_artifact.pk).exists()
        )
        self.assertTrue(Artifact.objects.filter(pk=source.pk).exists())

    def test_get_item_stats(self) -> None:
        source = self.playground.create_source_artifact()
        debian = Collection.objects.get(
            name="_", category=CollectionCategory.ARCHIVE
        )
        sid = self.playground.create_collection(
            "sid", category=CollectionCategory.SUITE
        )
        sid.manager.add_artifact(
            source,
            user=self.scenario.user,
            variables={"component": "main", "section": "extra"},
        )
        debian.manager.add_collection(sid, user=self.scenario.user)

        assert list(debian.get_item_stats()) == [
            {
                "type": CollectionItem.Types.COLLECTION,
                "category": "debian:suite",
                "active": True,
                "count": 1,
            }
        ]
        assert list(sid.get_item_stats()) == [
            {
                "type": CollectionItem.Types.ARTIFACT,
                "category": "debian:source-package",
                "active": True,
                "count": 1,
            }
        ]


class CollectionItemManagerTests(TestCase):
    """Tests for CollectionItemManager class."""

    user: ClassVar[User]
    workspace: ClassVar[Workspace]
    collection: ClassVar[Collection]
    workflow: ClassVar[WorkRequest]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.create_user(username="John")
        cls.workspace = cls.playground.create_workspace(name="System")
        cls.collection = Collection.objects.create(
            name="Name",
            category=CollectionCategory.ENVIRONMENTS,
            workspace=cls.workspace,
        )
        cls.workflow = cls.playground.create_worker_task(task_name="noop")

    def test_create_from_bare_data(self) -> None:
        """Verify create_from_bare_data method."""
        category = BareDataCategory.TEST
        name = "some-name"
        data = {"a": "b"}

        collection_item = CollectionItem.objects.create_from_bare_data(
            category,
            parent_collection=self.collection,
            name=name,
            data=data,
            created_by_user=self.user,
            created_by_workflow=self.workflow,
        )

        self.assertEqual(collection_item.parent_collection, self.collection)
        self.assertEqual(collection_item.name, name)
        self.assertEqual(collection_item.child_type, CollectionItem.Types.BARE)
        self.assertEqual(collection_item.category, category)
        self.assertEqual(collection_item.data, data)
        self.assertEqual(collection_item.created_by_user, self.user)
        self.assertEqual(collection_item.created_by_workflow, self.workflow)

    def test_create_from_bare_data_debusine_promise(self) -> None:
        """Verify create_from_bare_data with debusine:promise category."""
        category = BareDataCategory.PROMISE
        name = "valid-promise"
        data = DebusinePromise(
            promise_work_request_id=1,
            promise_workflow_id=100,
            promise_category=ArtifactCategory.BINARY_PACKAGE,
        )

        # Create a CollectionItem with valid DebusinePromiseData
        collection_item = CollectionItem.objects.create_from_bare_data(
            category,
            parent_collection=self.collection,
            name=name,
            data=data,
            created_by_user=self.user,
            created_by_workflow=self.workflow,
        )

        # Assert that the collection item is created correctly
        self.assertEqual(collection_item.parent_collection, self.collection)
        self.assertEqual(collection_item.name, name)
        self.assertEqual(collection_item.child_type, CollectionItem.Types.BARE)
        self.assertEqual(collection_item.category, category)
        self.assertEqual(collection_item.data, data.model_dump())
        self.assertEqual(collection_item.created_by_user, self.user)

    def test_create_from_bare_data_debusine_promise_invalid_data(self) -> None:
        """
        Verify create_from_bare_data raises ValueError.

        Invalid debusine:promise data.
        """
        category = BareDataCategory.PROMISE
        name = "invalid-promise"
        # Invalid data (missing required fields like promise_work_request_id)
        data = {
            "promise_workflow_id": 100,
            "promise_category": "debian:binary-package",
        }

        # Expect a ValueError due to missing promise_work_request_id in the data
        with self.assertRaises(ValueError):
            CollectionItem.objects.create_from_bare_data(
                category,
                parent_collection=self.collection,
                name=name,
                data=data,
                created_by_user=self.user,
                created_by_workflow=self.workflow,
            )

    def test_create_from_bare_data_created_at(self) -> None:
        created_at = timezone.now() - dt.timedelta(days=1)

        collection_item = CollectionItem.objects.create_from_bare_data(
            BareDataCategory.TEST,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
            created_at=created_at,
        )

        self.assertEqual(collection_item.created_at, created_at)
        self.assertIsNone(collection_item.removed_at)

    def test_create_from_bare_data_created_at_replaced_by(self) -> None:
        current_item = CollectionItem.objects.create_from_bare_data(
            BareDataCategory.TEST,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
        )
        created_at = timezone.now() - dt.timedelta(days=1)

        old_item = CollectionItem.objects.create_from_bare_data(
            BareDataCategory.TEST,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
            created_at=created_at,
            replaced_by=current_item,
        )

        self.assertEqual(old_item.created_at, created_at)
        self.assertEqual(old_item.removed_at, current_item.created_at)
        self.assertEqual(old_item.removed_by_user, current_item.created_by_user)
        self.assertEqual(
            old_item.removed_by_workflow, current_item.created_by_workflow
        )

    def test_create_from_artifact(self) -> None:
        """Verify create_from_artifact method."""
        data = {"a": "b"}
        artifact, _ = self.playground.create_artifact(data=data)

        name = "some-name"
        collection_item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=self.collection,
            name=name,
            data=data,
            created_by_user=self.user,
            created_by_workflow=self.workflow,
        )

        self.assertEqual(collection_item.parent_collection, self.collection)
        self.assertEqual(collection_item.name, name)
        self.assertEqual(collection_item.artifact, artifact)
        self.assertEqual(
            collection_item.child_type, CollectionItem.Types.ARTIFACT
        )
        self.assertEqual(collection_item.category, artifact.category)
        self.assertEqual(collection_item.data, data)
        self.assertEqual(collection_item.created_by_user, self.user)
        self.assertEqual(collection_item.created_by_workflow, self.workflow)

    def test_create_from_artifact_created_at(self) -> None:
        artifact, _ = self.playground.create_artifact()
        created_at = timezone.now() - dt.timedelta(days=1)

        collection_item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
            created_at=created_at,
        )

        self.assertEqual(collection_item.created_at, created_at)
        self.assertIsNone(collection_item.removed_at)

    def test_create_from_artifact_created_at_replaced_by(self) -> None:
        current_artifact, _ = self.playground.create_artifact()
        current_item = CollectionItem.objects.create_from_artifact(
            current_artifact,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
        )
        created_at = timezone.now() - dt.timedelta(days=1)

        old_artifact, _ = self.playground.create_artifact()
        old_item = CollectionItem.objects.create_from_artifact(
            old_artifact,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
            created_at=created_at,
            replaced_by=current_item,
        )

        self.assertEqual(old_item.created_at, created_at)
        self.assertEqual(old_item.removed_at, current_item.created_at)
        self.assertEqual(old_item.removed_by_user, current_item.created_by_user)
        self.assertEqual(
            old_item.removed_by_workflow, current_item.created_by_workflow
        )

    def test_create_from_collection(self) -> None:
        """Verify create_from_collection method."""
        category = "some-category"
        name = "collection-name"
        data = {"a": "b"}

        collection = Collection.objects.create(
            name="collection",
            category=category,
            workspace=self.workspace,
        )

        collection_item = CollectionItem.objects.create_from_collection(
            collection,
            parent_collection=self.collection,
            name=name,
            data=data,
            created_by_user=self.user,
            created_by_workflow=self.workflow,
        )

        self.assertEqual(collection_item.parent_collection, self.collection)
        self.assertEqual(collection_item.name, name)
        self.assertEqual(collection_item.collection, collection)
        self.assertEqual(
            collection_item.child_type, CollectionItem.Types.COLLECTION
        )
        self.assertEqual(collection_item.category, category)
        self.assertEqual(collection_item.data, data)
        self.assertEqual(collection_item.created_by_user, self.user)
        self.assertEqual(collection_item.created_by_workflow, self.workflow)

    def test_create_from_collection_created_at(self) -> None:
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        created_at = timezone.now() - dt.timedelta(days=1)

        collection_item = CollectionItem.objects.create_from_collection(
            collection,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
            created_at=created_at,
        )

        self.assertEqual(collection_item.created_at, created_at)
        self.assertIsNone(collection_item.removed_at)

    def test_create_from_collection_created_at_replaced_by(self) -> None:
        current_collection = self.playground.create_collection(
            "test-current", CollectionCategory.TEST
        )
        current_item = CollectionItem.objects.create_from_collection(
            current_collection,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
        )
        created_at = timezone.now() - dt.timedelta(days=1)

        old_collection = self.playground.create_collection(
            "test-old", CollectionCategory.TEST
        )
        old_item = CollectionItem.objects.create_from_collection(
            old_collection,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.user,
            created_by_workflow=self.workflow,
            created_at=created_at,
            replaced_by=current_item,
        )

        self.assertEqual(old_item.created_at, created_at)
        self.assertEqual(old_item.removed_at, current_item.created_at)
        self.assertEqual(old_item.removed_by_user, current_item.created_by_user)
        self.assertEqual(
            old_item.removed_by_workflow, current_item.created_by_workflow
        )

    def test_drop_full_history(self) -> None:
        """Verify drop_full_history method."""
        data = {"a": "b"}
        artifact, _ = self.playground.create_artifact(data=data)

        self.collection.full_history_retention_period = dt.timedelta(days=2)
        self.collection.save()

        collection_item_removed = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=3),
        )

        collection_item_young = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=1),
        )

        collection_item_keep = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
        )

        CollectionItem.objects.drop_full_history(timezone.now())

        collection_item_removed.refresh_from_db()
        self.assertIsNone(collection_item_removed.artifact)

        collection_item_young.refresh_from_db()
        self.assertIsNotNone(collection_item_young.artifact)

        collection_item_keep.refresh_from_db()
        self.assertIsNotNone(collection_item_keep.artifact)

    def test_drop_full_history_no_retention_period(self) -> None:
        """Verify drop_full_history method without retention_period."""
        data = {"a": "b"}
        artifact, _ = self.playground.create_artifact(data=data)

        collection_item_removed = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=3),
        )

        collection_item_young = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=1),
        )

        collection_item_keep = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
        )

        CollectionItem.objects.drop_full_history(timezone.now())

        collection_item_removed.refresh_from_db()
        self.assertIsNotNone(collection_item_removed.artifact)

        collection_item_young.refresh_from_db()
        self.assertIsNotNone(collection_item_young.artifact)

        collection_item_keep.refresh_from_db()
        self.assertIsNotNone(collection_item_keep.artifact)

    def test_drop_metadata(self) -> None:
        """Verify drop_metadata method."""
        data = {"a": "b"}
        artifact, _ = self.playground.create_artifact(data=data)

        self.collection.full_history_retention_period = dt.timedelta(days=1)
        self.collection.metadata_only_retention_period = dt.timedelta(days=1)
        self.collection.save()

        collection_item_removed = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=3),
        )

        collection_item_young = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=1),
        )

        collection_item_keep = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
        )

        CollectionItem.objects.drop_metadata(timezone.now())

        with self.assertRaises(CollectionItem.DoesNotExist):
            CollectionItem.objects.get(id=collection_item_removed.id)

        collection_item_young.refresh_from_db()
        self.assertIsNotNone(collection_item_young)

        collection_item_keep.refresh_from_db()
        self.assertIsNotNone(collection_item_keep)

    def test_drop_metadata_no_retention_period(self) -> None:
        """Verify drop_metadata method."""
        data = {"a": "b"}
        artifact, _ = self.playground.create_artifact(data=data)

        collection_item_removed = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=3),
        )

        collection_item_young = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
            removed_at=timezone.now() - dt.timedelta(days=1),
        )

        collection_item_keep = CollectionItem.objects.create(
            parent_collection=self.collection,
            name="some-name",
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=self.user,
        )

        CollectionItem.objects.drop_metadata(timezone.now())

        collection_item_removed.refresh_from_db()
        self.assertIsNotNone(collection_item_removed)

        collection_item_young.refresh_from_db()
        self.assertIsNotNone(collection_item_young)

        collection_item_keep.refresh_from_db()
        self.assertIsNotNone(collection_item_keep)

    def test_active_at(self) -> None:
        """``active_at`` filters to items active at that time."""
        items: list[CollectionItem] = []
        item = self.playground.create_debian_environment(
            collection=self.collection, codename="bookworm"
        )
        item.created_at = dt.datetime(2024, 12, 1, tzinfo=dt.UTC)
        item.save()
        items.append(item)
        for timestamp in (
            dt.datetime(2025, 1, 1, tzinfo=dt.UTC),
            dt.datetime(2025, 2, 1, tzinfo=dt.UTC),
            dt.datetime(2025, 3, 1, tzinfo=dt.UTC),
        ):
            self.collection.child_items.active().filter(
                data__codename="trixie"
            ).update(
                removed_by_user=self.playground.get_default_user(),
                removed_at=timestamp,
            )
            item = self.playground.create_debian_environment(
                collection=self.collection, codename="trixie"
            )
            item.created_at = timestamp
            item.save()
            items.append(item)
        self.collection.child_items.active().filter(
            data__codename="trixie"
        ).update(
            removed_by_user=self.playground.get_default_user(),
            removed_at=dt.datetime(2025, 4, 1, tzinfo=dt.UTC),
        )

        for timestamp, expected_items in (
            (dt.datetime(2024, 11, 30, tzinfo=dt.UTC), []),
            (dt.datetime(2024, 12, 1, tzinfo=dt.UTC), [items[0]]),
            (dt.datetime(2024, 12, 15, tzinfo=dt.UTC), [items[0]]),
            (
                dt.datetime(2025, 1, 1, tzinfo=dt.UTC),
                [items[0], items[1]],
            ),
            (
                dt.datetime(2025, 1, 15, tzinfo=dt.UTC),
                [items[0], items[1]],
            ),
            (
                dt.datetime(2025, 2, 1, tzinfo=dt.UTC),
                [items[0], items[2]],
            ),
            (
                dt.datetime(2025, 2, 15, tzinfo=dt.UTC),
                [items[0], items[2]],
            ),
            (
                dt.datetime(2025, 3, 1, tzinfo=dt.UTC),
                [items[0], items[3]],
            ),
            (
                dt.datetime(2025, 3, 15, tzinfo=dt.UTC),
                [items[0], items[3]],
            ),
            (dt.datetime(2025, 4, 1, tzinfo=dt.UTC), [items[0]]),
        ):
            self.assertQuerySetEqual(
                self.collection.child_items.active_at(timestamp),
                expected_items,
                ordered=False,
            )


class CollectionItemTests(TestCase):
    """Tests for CollectionItem class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.artifact, _ = self.playground.create_artifact()

        self.workspace = self.playground.create_workspace(name="System")
        self.collection = Collection.objects.create(
            name="Name",
            category=CollectionCategory.TEST,
            workspace=self.workspace,
        )
        self.user = get_user_model().objects.create_user(
            username="John", email="john@example.org"
        )

    def test_artifact_collection_item_same_name(self) -> None:
        """Two CollectionItem: same name, category different child_type."""
        name = "name:duplicated"
        category = "category:duplicated"

        CollectionItem.objects.create(
            name=name,
            category=category,
            parent_collection=self.collection,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=self.artifact,
            created_by_user=self.user,
        )
        with self.assertRaisesRegex(
            IntegrityError, "db_collectionitem_unique_active_name"
        ):
            CollectionItem.objects.create(
                name=name,
                category=category,
                parent_collection=self.collection,
                child_type=CollectionItem.Types.COLLECTION,
                collection=Collection.objects.create(
                    name="Test", category="test", workspace=self.workspace
                ),
                created_by_user=self.user,
            )

    def test_collection_and_parent_collection_not_the_same(self) -> None:
        """Cannot create CollectionItem with collection == parent_collection."""
        with self.assertRaisesRegex(
            IntegrityError, "db_collectionitem_distinct_parent_collection"
        ):
            CollectionItem.objects.create(
                name="Test",
                category="Category",
                parent_collection=self.collection,
                child_type=CollectionItem.Types.COLLECTION,
                collection=self.collection,
                created_by_user=self.user,
            )

    def test_only_one_active_item_in_collection(self) -> None:
        """Cannot create duplicated CollectionItem (not removed)."""
        name = "Name of the item"
        category = ArtifactCategory.SOURCE_PACKAGE

        CollectionItem.objects.create(
            name=name,
            category=category,
            artifact=self.artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        with self.assertRaisesRegex(
            IntegrityError,
            "db_collectionitem_unique_active_name",
        ):
            CollectionItem.objects.create(
                name=name,
                category=category,
                artifact=self.artifact,
                child_type=CollectionItem.Types.ARTIFACT,
                parent_collection=self.collection,
                created_by_user=self.user,
            )

    def test_second_active_item_added_first_one_removed(self) -> None:
        """Can create "duplicated" CollectionItem if first one is removed."""
        name = "Name of the item"
        category = ArtifactCategory.SOURCE_PACKAGE

        CollectionItem.objects.create(
            name=name,
            category=category,
            child_type=CollectionItem.Types.ARTIFACT,
            parent_collection=self.collection,
            created_by_user=self.user,
            removed_at=timezone.now(),
        )

        # Collection item can be added (no exception raised) because
        # the first one is removed
        CollectionItem.objects.create(
            name=name,
            category=category,
            artifact=self.artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

    def test_artifact_related_name(self) -> None:
        """Test artifact model "related_name" to CollectionItem."""
        artifact_in_collection_1 = CollectionItem.objects.create(
            name="name-1",
            category="category",
            artifact=self.artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        artifact_in_collection_2 = CollectionItem.objects.create(
            name="name-2",
            category="category",
            artifact=self.artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        self.assertQuerySetEqual(
            self.artifact.collection_items.all(),
            {artifact_in_collection_1, artifact_in_collection_2},
            ordered=False,
        )

    def test_artifact_constraint_are_valid_or_raise_integrity_error(
        self,
    ) -> None:
        """Verify constraints specific to an ARTIFACT type CollectionItem."""
        # CollectionItem of type ARTIFACT with an artifact
        CollectionItem.objects.create(
            name="Name",
            category="Category",
            artifact=self.artifact,
            parent_collection=self.collection,
            child_type=CollectionItem.Types.ARTIFACT,
            created_by_user=self.user,
        )

        # CollectionItem of type ARTIFACT without an artifact
        # (the artifact was removed)
        CollectionItem.objects.create(
            name="Name-2",
            category="Category-2",
            parent_collection=self.collection,
            child_type=CollectionItem.Types.ARTIFACT,
            created_by_user=self.user,
            removed_at=timezone.now(),
        )

        collection = Collection.objects.create(
            name="Collection-2", category="Category-2", workspace=self.workspace
        )

        # CollectionItem of type ARTIFACT with a collection
        with self.assertRaisesRegex(
            IntegrityError, "db_collectionitem_childtype_removedat_consistent"
        ):
            CollectionItem.objects.create(
                name="Name-3",
                category="Category-3",
                parent_collection=self.collection,
                child_type=CollectionItem.Types.ARTIFACT,
                collection=collection,
                created_by_user=self.user,
            )

    def test_collection_constraint_are_valid_or_raise_integrity_error(
        self,
    ) -> None:
        """Verify constraints specific to a COLLECTION type CollectionItem."""
        CollectionItem.objects.create(
            name="Name",
            category="Category",
            parent_collection=self.collection,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=self.artifact,
            removed_at=timezone.now(),
            created_by_user=self.user,
        )

        collection = Collection.objects.create(
            name="Collection-2", category="Category-2", workspace=self.workspace
        )

        CollectionItem.objects.create(
            name="Name-2",
            category="Category-2",
            parent_collection=self.collection,
            child_type=CollectionItem.Types.COLLECTION,
            collection=collection,
            removed_at=timezone.now(),
            created_by_user=self.user,
        )

    def test_collection_type_item_constraints(self) -> None:
        """Verify constraints specific to a Collection type CollectionItem."""
        collection = Collection.objects.create(
            name="Collection-2", category="Category-2", workspace=self.workspace
        )

        # CollectionItem of type COLLECTION with a collection
        CollectionItem.objects.create(
            name="Name",
            category="Category",
            collection=collection,
            parent_collection=self.collection,
            child_type=CollectionItem.Types.COLLECTION,
            created_by_user=self.user,
        )

        # CollectionItem of type COLLECTION without a collection (collection
        # was removed)
        CollectionItem.objects.create(
            name="Name-2",
            category="Category-2",
            parent_collection=self.collection,
            child_type=CollectionItem.Types.COLLECTION,
            created_by_user=self.user,
            removed_at=timezone.now(),
        )

        # CollectionItem of type COLLECTION with an artifact
        with self.assertRaisesRegex(
            IntegrityError, "db_collectionitem_childtype_removedat_consistent"
        ):
            CollectionItem.objects.create(
                name="Name-3",
                category="Category-3",
                parent_collection=self.collection,
                child_type=CollectionItem.Types.COLLECTION,
                artifact=self.artifact,
                created_by_user=self.user,
            )

    def test_bare_collection_item_constraints(self) -> None:
        """Verify constraints specific to a Bare type CollectionItem."""
        CollectionItem.objects.create(
            name="Name",
            category="Bare",
            parent_collection=self.collection,
            child_type=CollectionItem.Types.BARE,
            created_by_user=self.user,
        )

        with self.assertRaisesRegex(
            IntegrityError, "db_collectionitem_childtype_removedat_consistent"
        ):
            CollectionItem.objects.create(
                name="Name",
                category="Category",
                artifact=self.artifact,
                parent_collection=self.collection,
                child_type=CollectionItem.Types.BARE,
                created_by_user=self.user,
            )

    def test_debian_environments_no_more_than_one_codename_architecture(
        self,
    ) -> None:
        """Cannot create more than one duplicated active debian:environments."""
        collection = Collection.objects.create(
            name="Name",
            category=CollectionCategory.ENVIRONMENTS,
            workspace=self.workspace,
        )
        category = ArtifactCategory.SYSTEM_TARBALL
        data = {"codename": "bookworm", "architecture": "amd64"}

        artifact_1, _ = self.playground.create_artifact(category=category)
        CollectionItem.objects.create(
            name="bookworm-amd64",
            category=category,
            parent_collection=collection,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact_1,
            created_by_user=self.user,
            data=data,
            removed_at=timezone.now(),
        )

        # Can be created because the previous one is removed
        artifact_2, _ = self.playground.create_artifact(category=category)
        CollectionItem.objects.create(
            name="bookworm-amd64-2",
            category=category,
            parent_collection=collection,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact_2,
            created_by_user=self.user,
            data=data,
        )

        # Can be created because it's different variant
        artifact_3, _ = self.playground.create_artifact(category=category)
        CollectionItem.objects.create(
            name="bookworm-amd64-3",
            category=category,
            parent_collection=collection,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact_3,
            created_by_user=self.user,
            data={**data, "variant": "buildd"},
        )

        # Can be created because it's different backend
        artifact_4, _ = self.playground.create_artifact(category=category)
        CollectionItem.objects.create(
            name="bookworm-amd64-4",
            category=category,
            parent_collection=collection,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact_4,
            created_by_user=self.user,
            data={**data, "backend": BackendType.UNSHARE},
        )

        # Can be created: duplicated but in a different collection
        artifact_5, _ = self.playground.create_artifact(category=category)
        CollectionItem.objects.create(
            name="bookworm-amd64-5",
            category=category,
            parent_collection=Collection.objects.create(
                name="Testing",
                category=CollectionCategory.ENVIRONMENTS,
                workspace=self.workspace,
            ),
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact_5,
            created_by_user=self.user,
            data={**data, "variant": "buildd"},
        )

        # Cannot be created because already one active artifact
        # with the same data in the same collection (from artifact_3).
        artifact_6, _ = self.playground.create_artifact(category=category)
        msg = "db_collectionitem_unique_debian_environment"
        with self.assertRaisesRegex(IntegrityError, msg):
            CollectionItem.objects.create(
                name="bookworm-amd64-6",
                category=category,
                parent_collection=collection,
                child_type=CollectionItem.Types.ARTIFACT,
                artifact=artifact_6,
                created_by_user=self.user,
                data={**data, "variant": "buildd"},
            )

    def test_str_collection_item_artifact(self) -> None:
        """Stringification is correct for CollectionItem for ARTIFACT."""
        name = "Name of the item"
        category = ArtifactCategory.SOURCE_PACKAGE
        artifact_item = CollectionItem.objects.create(
            name=name,
            category=category,
            artifact=self.artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            parent_collection=self.collection,
            created_by_user=self.user,
        )
        self.assertEqual(
            str(artifact_item),
            f"Id: {artifact_item.id} Name: {name} "
            f"Parent collection id: {self.collection.id} "
            f"Child type: {artifact_item.child_type} "
            f"Artifact id: {self.artifact.id}",
        )

    def test_str_collection_item_collection(self) -> None:
        """Stringification contains collection id.."""
        collection = Collection.objects.create(
            name="Collection-2", category="Category-2", workspace=self.workspace
        )

        collection_item = CollectionItem.objects.create(
            name="Name of the item",
            category=ArtifactCategory.SOURCE_PACKAGE,
            collection=collection,
            child_type=CollectionItem.Types.COLLECTION,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        self.assertIn(f"Collection id: {collection.id}", str(collection_item))

    def test_str_collection_item_bare(self) -> None:
        """Stringification does not contain collection/artifact id."""
        collection_item = CollectionItem.objects.create(
            name="Name of the item",
            category=ArtifactCategory.SOURCE_PACKAGE,
            child_type=CollectionItem.Types.BARE,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        self.assertNotIn("Collection id:", str(collection_item))
        self.assertNotIn("Artifact id:", str(collection_item))

    def test_parent_category(self) -> None:
        """`parent_category` is copied from the parent collection."""
        collection_item = CollectionItem.objects.create(
            name="test",
            category=BareDataCategory.TEST,
            child_type=CollectionItem.Types.BARE,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        self.assertEqual(
            collection_item.parent_category, self.collection.category
        )

    def test_ui(self) -> None:
        from debusine.web.views.ui.collections import CollectionItemUI

        item = CollectionItem.objects.create(
            name="Name of the item",
            category=ArtifactCategory.SOURCE_PACKAGE,
            child_type=CollectionItem.Types.BARE,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        request = RequestFactory().get("/")
        ui = item.ui(request)
        self.assertIsInstance(ui, CollectionItemUI)

    def test_get_absolute_url(self) -> None:
        """Test get_absolute_url results."""
        item = CollectionItem.objects.create(
            name="Name of the item",
            category=ArtifactCategory.SOURCE_PACKAGE,
            child_type=CollectionItem.Types.BARE,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        item_url = item.get_absolute_url()
        self.assertEqual(
            item_url,
            f"/{self.workspace.scope.name}/{self.workspace.name}"
            f"/collection/{self.collection.category}/{self.collection.name}"
            f"/item/{item.pk}/{quote(item.name)}/",
        )
        self.assertTrue(item_url.endswith("/Name%20of%20the%20item/"))

    def test_get_absolute_url_slash(self) -> None:
        """Test get_absolute_url results when the item name contains `/`."""
        item = CollectionItem.objects.create(
            name="some/hierarchical/name",
            category=ArtifactCategory.SOURCE_PACKAGE,
            child_type=CollectionItem.Types.BARE,
            parent_collection=self.collection,
            created_by_user=self.user,
        )

        item_url = item.get_absolute_url()
        self.assertEqual(
            item_url,
            f"/{self.workspace.scope.name}/{self.workspace.name}"
            f"/collection/{self.collection.category}/{self.collection.name}"
            f"/item/{item.pk}/some/hierarchical/name/",
        )

    def test_get_absolute_url_different_scope(self) -> None:
        """get_absolute_url works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        collection = self.playground.create_collection(
            "Name", CollectionCategory.TEST, workspace=other_workspace
        )
        item = CollectionItem.objects.create(
            name="Name of the item",
            category=ArtifactCategory.SOURCE_PACKAGE,
            child_type=CollectionItem.Types.BARE,
            parent_collection=collection,
            created_by_user=self.user,
        )

        item_url = item.get_absolute_url()
        self.assertEqual(
            item_url,
            f"/{other_scope.name}/{other_workspace.name}"
            f"/collection/{collection.category}/{collection.name}"
            f"/item/{item.pk}/{quote(item.name)}/",
        )
        self.assertTrue(item_url.endswith("/Name%20of%20the%20item/"))

    def test_expand_variables(self) -> None:
        """Variables are correctly expanded, with error handling."""
        variables = {
            "$package": "deb_fields.Package",
            "$version": "deb_fields.Version",
            "constant": "value",
        }
        reference_data = {
            "deb_fields": {
                "Package": "hello",
                "Version": "2.10-3",
                "Architecture": "any",
            }
        }

        self.assertEqual(
            CollectionItem.expand_variables(variables, reference_data),
            {"package": "hello", "version": "2.10-3", "constant": "value"},
        )

        variables = {
            "$package": "deb_fields.*",
            "$version": "deb_fields.Version",
        }
        with self.assertRaisesRegex(ValueError, "Too many values expanding"):
            CollectionItem.expand_variables(variables, reference_data)

        variables = {
            "$package": "]",
        }
        with self.assertRaisesRegex(ValueError, "Parse error"):
            CollectionItem.expand_variables(variables, reference_data)

        variables = {"$package": "deb_fields.Package", "package": "constant"}
        with self.assertRaisesRegex(
            ValueError, r"Cannot set both '\$package' and 'package' variables"
        ):
            CollectionItem.expand_variables(variables, reference_data)

        variables = {"$version": "deb_fields.Version"}
        with self.assertRaises(KeyError):
            CollectionItem.expand_variables(variables, {})

    def test_expand_name(self) -> None:
        """Item name is correctly expanded."""
        item_template = "{package}_{version}"
        expanded_variables = {"package": "hello", "version": "2.10-3"}

        self.assertEqual(
            CollectionItem.expand_name(item_template, expanded_variables),
            "hello_2.10-3",
        )


@dataclass(slots=True)
class Suites:
    """Suites used by the tests."""

    trixie: Collection
    trixie_security: Collection
    trixie_backports: Collection
    bookworm: Collection
    personal_suite: Collection


class CollectionRelationTests(TestCase):
    """Tests for CollectionRelation class."""

    workspace: Workspace
    suites: Suites
    qa_results: Collection
    user: User

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.workspace = cls.playground.create_workspace()
        cls.suites = Suites(
            trixie=cls.playground.create_collection(
                name="trixie", category=CollectionCategory.SUITE
            ),
            trixie_security=cls.playground.create_collection(
                name="trixie-security", category=CollectionCategory.SUITE
            ),
            trixie_backports=cls.playground.create_collection(
                name="trixie-backports", category=CollectionCategory.SUITE
            ),
            bookworm=cls.playground.create_collection(
                name="bookworm", category=CollectionCategory.SUITE
            ),
            personal_suite=cls.playground.create_collection(
                name="personal", category=CollectionCategory.SUITE
            ),
        )

        cls.qa_results = cls.playground.create_collection(
            name="trixie-qa", category=CollectionCategory.QA_RESULTS
        )
        cls.user = cls.playground.create_user("foo")

    def create(self, **kwargs: Any) -> CollectionRelation:
        """Create CollectionRelation with **kwargs and a user."""
        return CollectionRelation.objects.create(
            **kwargs,
            created_by_user=self.user,
        )

    def test_str_with_position(self) -> None:
        relation = self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_backports,
            position=1,
        )

        self.assertEqual(
            str(relation), "trixie → trixie-backports (Requires, position 1)"
        )

    def test_str_without_position(self) -> None:
        relation = self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
            target=self.qa_results,
        )

        self.assertEqual(
            str(relation), "trixie → trixie-qa (Default QA results)"
        )

    def test_suite_requires_without_position_raises(self) -> None:
        with self.assertRaisesRegex(
            IntegrityError, "db_collectionrelation_position"
        ):
            self.create(
                source=self.suites.trixie,
                type=CollectionRelation.Types.SUITE_REQUIRES,
                target=self.suites.trixie_backports,
            )

    def test_suite_requires_accept_distinct_positions(self) -> None:
        self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_backports,
            position=1,
        )
        self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_security,
            position=2,
        )

    def test_suite_requires_no_repeated_target(self) -> None:
        self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_backports,
            position=1,
        )

        with self.assertRaisesRegex(
            IntegrityError,
            "db_collectionrelation_suite_requires_no_duplicate_target",
        ):
            self.create(
                source=self.suites.trixie,
                type=CollectionRelation.Types.SUITE_REQUIRES,
                target=self.suites.trixie_backports,
                position=2,
            )

    def test_no_duplicate_position(self) -> None:
        self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_security,
            position=1,
        )

        second = self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_backports,
            position=2,
        )

        CollectionRelation.objects.filter(pk=second.pk).update(position=1)
        with self.assertRaisesRegex(
            IntegrityError,
            "db_collectionrelation_suite_requires_no_duplicate_position",
        ):
            connection.check_constraints()

    def test_start_position_1(self) -> None:
        relation = self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie,
            position=-4,
        )

        self.assertEqual(relation.position, 1)

    def test_maximum_position_current_max_1(self) -> None:
        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie,
            position=1,
        )
        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_security,
            position=2,
        )
        backports = self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_backports,
            position=5,
        )

        self.assertEqual(backports.position, 3)

    def test_check_cardinality_max_1(self) -> None:
        class RelationParameters(NamedTuple):
            source: Collection
            type: CollectionRelation.Types
            target: Collection

        test_cases = [
            RelationParameters(
                source=self.suites.personal_suite,
                type=CollectionRelation.Types.SUITE_FORKED_FROM,
                target=self.suites.trixie,
            ),
            RelationParameters(
                source=self.suites.personal_suite,
                type=CollectionRelation.Types.SUITE_BASED_ON,
                target=self.suites.trixie,
            ),
            RelationParameters(
                source=self.suites.personal_suite,
                type=CollectionRelation.Types.SUITE_TARGETING,
                target=self.suites.trixie,
            ),
            RelationParameters(
                source=self.suites.personal_suite,
                type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
                target=self.qa_results,
            ),
        ]

        for test_case in test_cases:
            with self.subTest(type=test_case.type.value):
                self.create(**test_case._asdict())

                with (
                    transaction.atomic(),
                    self.assertRaisesRegex(
                        IntegrityError,
                        "db_collectionrelation_cardinality_checks_max_1",
                    ),
                ):
                    self.create(**test_case._asdict())

    def test_can_add_multiple_targets(self) -> None:
        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_FORKED_FROM,
            target=self.suites.trixie,
        )

        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            target=self.suites.trixie,
        )

        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_TARGETING,
            target=self.suites.trixie,
        )

        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
            target=self.qa_results,
        )

        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.bookworm,
            position=1,
        )

        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie,
            position=2,
        )

    def test_created_by_user_mandatory(self) -> None:
        with self.assertRaisesRegex(IntegrityError, "created_by_user_id"):
            CollectionRelation.objects.create(
                source=self.suites.trixie,
                type=CollectionRelation.Types.SUITE_BASED_ON,
                target=self.suites.bookworm,
            )

    def test_created_updated_audit(self) -> None:
        relation = self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            target=self.suites.bookworm,
        )
        # Assert "created_at" and "updated_at"
        self.assertIsNotNone(relation.created_at)
        self.assertIsNone(relation.updated_at)

        self.assertEqual(relation.created_by_user, self.user)
        self.assertIsNone(relation.updated_by_user)

        # Update relation
        updated_by_user = self.playground.create_user("bar")
        relation.target = self.suites.trixie_backports

        relation.save(updated_by_user=updated_by_user)
        relation.refresh_from_db()

        # Assert that "updated_by_user" and "updated_at" fields are updated
        self.assertEqual(relation.updated_by_user, updated_by_user)
        assert relation.updated_at is not None
        self.assertGreater(relation.updated_at, relation.created_at)

    def test_save_updated_by_user_mandatory(self) -> None:
        relation = self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            target=self.suites.bookworm,
        )

        relation.target = self.suites.trixie_backports
        with self.assertRaisesRegex(
            ValueError,
            "updated_by_user must be set when updating a CollectionRelation",
        ):
            relation.save()

    def test_save_update_fields_not_supported(self) -> None:
        relation = self.create(
            source=self.suites.trixie,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            target=self.suites.bookworm,
        )

        with self.assertRaisesRegex(
            ValueError,
            "update_fields is not supported when updating CollectionRelation",
        ):
            relation.save(update_fields=["target"], updated_by_user=self.user)

    def test_types_without_position_reject_position(self) -> None:
        class RelationParameters(NamedTuple):
            source: Collection
            type: CollectionRelation.Types
            target: Collection

        test_cases = [
            RelationParameters(
                source=self.suites.personal_suite,
                type=CollectionRelation.Types.SUITE_FORKED_FROM,
                target=self.suites.trixie,
            ),
            RelationParameters(
                source=self.suites.personal_suite,
                type=CollectionRelation.Types.SUITE_BASED_ON,
                target=self.suites.trixie,
            ),
            RelationParameters(
                source=self.suites.personal_suite,
                type=CollectionRelation.Types.SUITE_TARGETING,
                target=self.suites.trixie,
            ),
            RelationParameters(
                source=self.suites.trixie,
                type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
                target=self.qa_results,
            ),
        ]

        for test_case in test_cases:
            with (
                self.subTest(type=test_case.type.value),
                transaction.atomic(),
                self.assertRaisesRegex(
                    IntegrityError, "db_collectionrelation_position"
                ),
            ):
                self.create(**test_case._asdict(), position=1)

    def test_collection_categories_match_expected(
        self,
    ) -> None:
        class RelationParameters(NamedTuple):
            source: Collection
            source_category: CollectionCategory
            type: CollectionRelation.Types
            target: Collection
            target_category: CollectionCategory

        # To keep it short, this is not covering all the invalid cases
        test_cases = [
            RelationParameters(
                source=self.suites.trixie,
                source_category=CollectionCategory.ENVIRONMENTS,  # invalid
                type=CollectionRelation.Types.SUITE_FORKED_FROM,
                target=self.suites.bookworm,
                target_category=CollectionCategory.SUITE,
            ),
            RelationParameters(
                source=self.suites.trixie,
                source_category=CollectionCategory.SUITE,
                type=CollectionRelation.Types.SUITE_BASED_ON,
                target=self.suites.bookworm,
                target_category=CollectionCategory.ENVIRONMENTS,  # invalid
            ),
            RelationParameters(
                source=self.suites.trixie,
                source_category=CollectionCategory.SUITE,
                type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
                target=self.qa_results,
                target_category=CollectionCategory.SUITE,  # invalid
            ),
        ]

        for test_case in test_cases:
            with (
                self.subTest(type=test_case.type.value),
                transaction.atomic(),
                self.assertRaisesRegex(
                    IntegrityError, "db_collectionrelation_categories_correct"
                ),
            ):
                CollectionRelation.objects.bulk_create(
                    [
                        CollectionRelation(
                            **test_case._asdict(),
                            created_by_user=self.user,
                            updated_by_user=self.user,
                        )
                    ]
                )

    def test_create_populates_categories(self) -> None:
        relation = self.create(
            source=self.suites.trixie,
            target=self.qa_results,
            type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
        )

        self.assertEqual(relation.source_category, self.suites.trixie.category)
        self.assertEqual(relation.target_category, self.qa_results.category)

    def test_delete(self) -> None:
        relation = self.create(
            source=self.suites.trixie,
            target=self.qa_results,
            type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
        )

        relation_id = relation.id

        relation.delete()

        self.assertFalse(
            CollectionRelation.objects.filter(id=relation_id).exists()
        )

    def test_delete_shift_positions(self) -> None:
        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.bookworm,
            position=1,
        )

        to_delete = self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie,
            position=2,
        )

        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_backports,
            position=3,
        )

        self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.trixie_security,
            position=4,
        )

        to_delete.delete()

        self.assertEqual(
            list(
                CollectionRelation.objects.filter(
                    source=self.suites.personal_suite,
                    type=CollectionRelation.Types.SUITE_REQUIRES,
                )
                .order_by("position")
                .values_list("target__name", "position")
            ),
            [
                (self.suites.bookworm.name, 1),
                (self.suites.trixie_backports.name, 2),
                (self.suites.trixie_security.name, 3),
            ],
        )

    def test_can_delete(self) -> None:
        """CollectionRelation.can_delete delegates to source collection."""
        relation = self.create(
            source=self.suites.personal_suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            target=self.suites.bookworm,
            position=1,
        )

        pc = PermissionContext.for_ui_user(self.user)

        with override_permission(Collection, "can_configure", AllowAll):
            self.assertTrue(relation.can_delete(pc))

        with override_permission(Collection, "can_configure", DenyAll):
            self.assertFalse(relation.can_delete(pc))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for signon provider backends."""

import contextlib
import io
import json
from collections.abc import Generator, Sequence
from types import SimpleNamespace
from typing import ClassVar, override
from unittest import mock

import django.http
import requests
from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory, TestCase, override_settings
from jwcrypto import jwk, jws, jwt

from debusine.server.signon import providers
from debusine.server.signon.tests.test_signon import MockSession

SALSA_PROVIDER = providers.GitlabProvider(
    name="salsa",
    label="Salsa",
    icon="signon/gitlabian.svg",
    client_id="123client_id",
    client_secret="123client_secret",
    url="https://salsa.debian.org",
    scope=("openid", "profile", "email"),
)

GITLAB_PROVIDER = providers.GitlabProvider(
    name="gitlab",
    label="GitLab",
    icon="signon/gitlab.svg",
    client_id="123client_id",
    client_secret="123client_secret",
    url="https://gitlab.com",
)


class TestProvider(TestCase):
    """Test the base Provider class."""

    def test_defaults(self) -> None:
        """Test instantiating Provider with default arguments."""
        p = providers.Provider("name", "label")
        self.assertEqual(p.name, "name")
        self.assertEqual(p.label, "label")
        self.assertIsNone(p.icon)
        self.assertEqual(p.options, {})

    def test_all_set(self) -> None:
        """Test instantiating Provider."""
        p = providers.Provider(
            "name", "label", icon="icon", options={"answer": 42}
        )
        self.assertEqual(p.name, "name")
        self.assertEqual(p.label, "label")
        self.assertEqual(p.icon, "icon")
        self.assertEqual(p.options, {"answer": 42})


class SignonProviders(TestCase):
    """Test signon provider definitions."""

    remote_key: ClassVar[jwk.JWK]
    remote_keyset: ClassVar[jwk.JWKSet]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        """Create a reusable remote keyset only once for all the tests."""
        super().setUpClass()
        # Keyset used for mocking encrypted communication with the OIDC
        # provider.
        #
        # This would normally use an asymmetric key, but we use a symmetric one
        # to avoid slowing down tests.
        cls.remote_key = jwk.JWK.generate(kty='oct', size=256)
        cls.remote_keyset = jwk.JWKSet.from_json(
            json.dumps(
                {
                    "keys": [cls.remote_key.export_symmetric(as_dict=True)],
                }
            )
        )

    @contextlib.contextmanager
    def mock_load_keyset(
        self, provider: providers.OIDCProvider
    ) -> Generator[None]:
        """Do the right mocking setup to prepare loading a remote keyset."""
        # Mock answer with the keyset
        response = requests.Response()
        response.url = provider.url_jwks
        response.status_code = 200
        # Mock fetching the remote keyset
        response.raw = io.BytesIO(self.remote_keyset.export().encode())

        # On first access it fetches remote keys
        with mock.patch(
            "requests_oauthlib.OAuth2Session.get", return_value=response
        ):
            yield

    @contextlib.contextmanager
    def mock_fetch_token(
        self, response_token: str, time: int = 123450
    ) -> Generator[None]:
        """Mock the remote call in OAuth2Session.fetch_token."""
        with (
            mock.patch(
                "requests_oauthlib.OAuth2Session.fetch_token",
                return_value={
                    "id_token": response_token,
                    "access_token": "accesstoken",
                },
            ),
            mock.patch(
                "time.time",
                return_value=time,
            ),
        ):
            yield

    @contextlib.contextmanager
    def mock_load_tokens(
        self,
        provider: providers.OIDCProvider,
        response_token: str,
        time: int = 123450,
    ) -> Generator[None]:
        """Do the mocking needed to call load_tokens."""
        with (
            self.mock_load_keyset(provider),
            self.mock_fetch_token(response_token, time),
        ):
            yield

    def _make_request(
        self,
        session_state: str | None = "teststate",
        remote_state: str | None = "teststate",
        session_options: Sequence[str] | None = (),
    ) -> django.http.HttpRequest:
        """Create a test request."""
        # Mock request
        request_factory = RequestFactory()
        data = {}
        if remote_state is not None:
            data["state"] = remote_state
        request = request_factory.get("/callback", data=data)
        request.session = MockSession()
        if session_state is not None:
            request.session["signon_state_salsa"] = session_state
        if session_options is not None:
            request.session["signon_state_salsa_options"] = session_options
        return request

    def assertSignonStateRemoved(
        self, request: django.http.HttpRequest
    ) -> None:
        """Ensure that the signon state has been removed from session."""
        self.assertNotIn("signon_state_salsa", request.session)
        self.assertNotIn("signon_state_salsa_options", request.session)

    def test_get_setting_undefined(self) -> None:
        """Lookup with no SIGNON_PROVIDERS raises ImproperlyConfigured."""
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            "signon provider salsa requested, but SIGNON_PROVIDERS is not "
            "defined in settings",
        ):
            providers.get("salsa")

    @override_settings(SIGNON_PROVIDERS=[])
    def test_get_empty(self) -> None:
        """Lookup in an empty provider list raises ImproperlyConfigured."""
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            "signon provider salsa requested, but not found in "
            "SIGNON_PROVIDERS setting",
        ):
            providers.get("salsa")

    @override_settings(SIGNON_PROVIDERS=[SimpleNamespace(name="salsa")])
    def test_get_wrong_type(self) -> None:
        """A provider of the wrong type raises ImproperlyConfigured."""
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            "signon provider salsa requested, but its entry in "
            "SIGNON_PROVIDERS setting is not a Provider",
        ):
            providers.get("salsa")

    @override_settings(SIGNON_PROVIDERS=[SALSA_PROVIDER, GITLAB_PROVIDER])
    def test_get(self) -> None:
        """Test GitLab provider parameter construction."""
        with self.subTest(provider="salsa"):
            salsa = providers.get("salsa")
            self.assertIsNotNone(salsa)

            # Getting a provider twice returns the same object
            self.assertIs(providers.get("salsa"), salsa)

            assert isinstance(salsa, providers.GitlabProvider)
            self.assertEqual(salsa.name, "salsa")
            self.assertEqual(salsa.label, "Salsa")
            self.assertEqual(salsa.icon, "signon/gitlabian.svg")
            self.assertEqual(salsa.scope, ["openid", "profile", "email"])
            self.assertEqual(salsa.client_id, "123client_id")
            self.assertEqual(salsa.client_secret, "123client_secret")
            self.assertEqual(salsa.url_issuer, "https://salsa.debian.org")
            self.assertEqual(
                salsa.url_authorize, "https://salsa.debian.org/oauth/authorize"
            )
            self.assertEqual(
                salsa.url_token, "https://salsa.debian.org/oauth/token"
            )
            self.assertEqual(
                salsa.url_userinfo, "https://salsa.debian.org/oauth/userinfo"
            )
            self.assertEqual(
                salsa.url_jwks, "https://salsa.debian.org/oauth/discovery/keys"
            )

        with self.subTest(provider="gitlab"):
            gitlab = providers.get("gitlab")
            self.assertIsNotNone(gitlab)

            assert isinstance(gitlab, providers.GitlabProvider)
            self.assertEqual(gitlab.name, "gitlab")
            self.assertEqual(gitlab.label, "GitLab")
            self.assertEqual(gitlab.icon, "signon/gitlab.svg")
            self.assertEqual(gitlab.scope, ["openid"])
            self.assertEqual(gitlab.client_id, "123client_id")
            self.assertEqual(gitlab.client_secret, "123client_secret")
            self.assertEqual(gitlab.url_issuer, "https://gitlab.com")
            self.assertEqual(
                gitlab.url_authorize, "https://gitlab.com/oauth/authorize"
            )
            self.assertEqual(gitlab.url_token, "https://gitlab.com/oauth/token")
            self.assertEqual(
                gitlab.url_userinfo, "https://gitlab.com/oauth/userinfo"
            )
            self.assertEqual(
                gitlab.url_jwks, "https://gitlab.com/oauth/discovery/keys"
            )

    def test_entra_url_generation(self) -> None:
        tenant_id = "ca87cd01-56be-4303-a6c9-a10a6440ec4d"
        entra = providers.AzureEntraProvider(
            name="entra",
            label="Azure Entra",
            client_id="123client_id",
            client_secret="123client_secret",
            tenant_id=tenant_id,
        )

        self.assertEqual(entra.name, "entra")
        self.assertEqual(entra.label, "Azure Entra")
        self.assertEqual(entra.scope, ["openid", "email", "profile"])
        self.assertEqual(entra.client_id, "123client_id")
        self.assertEqual(entra.client_secret, "123client_secret")

        self.assertEqual(
            entra.url_issuer,
            f"https://login.microsoftonline.com/{tenant_id}/v2.0",
        )
        self.assertEqual(
            entra.url_authorize,
            (
                f"https://login.microsoftonline.com/{tenant_id}/oauth2/"
                f"v2.0/authorize"
            ),
        )
        self.assertEqual(
            entra.url_token,
            f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
        )
        self.assertEqual(
            entra.url_userinfo, "https://graph.microsoft.com/oidc/userinfo"
        )
        self.assertEqual(
            entra.url_jwks,
            f"https://login.microsoftonline.com/{tenant_id}/discovery/v2.0/keys",
        )

    def test_bind(self) -> None:
        """Provider.bind binds to the request and proxies correctly."""
        request = self._make_request()
        provider = providers.GitlabProvider(
            name="salsa",
            label="Salsa",
            client_id="123client_id",
            client_secret="123client_secret",
            url="https://salsa.debian.org",
            scope=("openid", "profile", "email"),
        )
        bound = provider.bind(request)
        self.assertIs(bound.provider, provider)
        self.assertIs(bound.request, request)
        self.assertEqual(bound.name, provider.name)
        self.assertEqual(bound.label, provider.label)
        self.assertEqual(bound.client_id, provider.client_id)
        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)

    def test_oidc_load_keys(self) -> None:
        """OIDC keys loads and parses the keyset."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request()
        bound = provider.bind(request)

        # On first access it fetches remote keys
        with self.mock_load_keyset(provider):
            self.assertIsNotNone(bound.keyset)

        # Subsequent accesses have the results cached
        with mock.patch("requests_oauthlib.OAuth2Session.get") as session_get:
            self.assertIsNotNone(bound.keyset)
        session_get.assert_not_called()

    def _make_response_token(
        self,
        provider: providers.OIDCProvider,
        iss: str | None = None,
        aud: str | None = None,
        exp: int = 123456,
        key: jwk.JWK | None = None,
    ) -> str:
        """Generate a response token."""
        if iss is None:
            iss = provider.url_issuer
        if aud is None:
            aud = provider.client_id

        # Encrypted response payload from the OIDC authentication provider
        response_payload = jwt.JWT(
            header={"alg": "HS256"},
            claims={
                "iss": iss,
                "aud": aud,
                "sub": "1234",
                "profile": "https://salsa.debian.org/username",
                "exp": exp,
            },
        )
        if key is None:
            key = self.remote_key
        response_payload.make_signed_token(key)
        response_token = response_payload.serialize()
        assert isinstance(response_token, str)
        return response_token

    def test_oidc_load_tokens(self) -> None:
        """Tokens are loaded from the auth provider and decoded."""
        provider = SALSA_PROVIDER
        request = self._make_request()
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)

        with self.mock_load_tokens(provider, response_token):
            bound.load_tokens()

        self.assertIsNotNone(bound.tokens)
        self.assertIsNotNone(bound.id_token_claims)
        self.assertEqual(bound.options, ())
        self.assertSignonStateRemoved(request)

    def test_oidc_load_tokens_options(self) -> None:
        """Tokens are loaded from the auth provider and decoded."""
        provider = SALSA_PROVIDER
        request = self._make_request(session_options=["bind"])
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)

        with self.mock_load_tokens(provider, response_token):
            bound.load_tokens()

        self.assertIsNotNone(bound.tokens)
        self.assertIsNotNone(bound.id_token_claims)
        self.assertEqual(bound.options, ["bind"])
        self.assertSignonStateRemoved(request)

    def test_oidc_load_state_not_in_session(self) -> None:
        """Callback state argument must match session."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request(session_state=None)
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)

        with (
            self.assertRaisesRegex(
                providers.OIDCValidationError,
                r"expected state not found in session",
            ),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_options_not_in_session(self) -> None:
        """Callback options argument must be in session."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request(session_options=None)
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)

        with (
            self.assertRaisesRegex(
                providers.OIDCValidationError,
                "options not found in session",
            ),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_state_not_in_get(self) -> None:
        """Callback state argument must match session."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request(remote_state=None)
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)

        with (
            self.assertRaisesRegex(
                providers.OIDCValidationError,
                r"state not found in remote response",
            ),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_wrong_state(self) -> None:
        """Callback state argument must match session."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request(session_state="wrong")
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)

        with (
            self.assertRaisesRegex(
                providers.OIDCValidationError, r"Request state mismatch"
            ),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_expired_token(self) -> None:
        """An expired token is rejected."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request()
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)

        with (
            self.assertRaises(jwt.JWTExpired),
            self.mock_load_tokens(provider, response_token, time=123556),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_tokens_wrong_keys(self) -> None:
        """Remote tokens need to be signed with the right keys."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request()
        bound = provider.bind(request)

        wrong_key = jwk.JWK.generate(kty='oct', size=256)
        response_token = self._make_response_token(provider, key=wrong_key)

        with (
            self.assertRaises(jwt.JWTMissingKey),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_tokens_wrong_issuer(self) -> None:
        """Remote tokens need to have the right issuer claim."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request()
        bound = provider.bind(request)
        response_token = self._make_response_token(provider, iss="WRONG")

        with (
            self.assertRaisesRegex(
                providers.OIDCValidationError, r"Issuer mismatch"
            ),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_tokens_wrong_audience(self) -> None:
        """Remote tokens need to have the right audience claim."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request()
        bound = provider.bind(request)
        response_token = self._make_response_token(provider, aud="WRONG")

        with (
            self.assertRaisesRegex(
                providers.OIDCValidationError, r"Audience mismatch"
            ),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)

    def test_oidc_load_tokens_corrupted(self) -> None:
        """A remote tokens that is corrupted is detected."""
        # Mock provider
        provider = SALSA_PROVIDER
        request = self._make_request()
        bound = provider.bind(request)
        response_token = self._make_response_token(provider)
        response_token = response_token[:4] + '0' + response_token[5:]

        with (
            self.assertRaises(jws.InvalidJWSObject),
            self.mock_load_tokens(provider, response_token),
        ):
            bound.load_tokens()

        self.assertIsNone(bound.tokens)
        self.assertIsNone(bound.id_token_claims)
        self.assertSignonStateRemoved(request)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Interface for executor backends."""

import os
from abc import ABC, abstractmethod
from enum import StrEnum
from pathlib import Path, PurePath
from stat import S_IMODE
from subprocess import CalledProcessError, CompletedProcess, PIPE
from typing import Any, AnyStr, ClassVar, Literal, Optional, overload, override

from debusine.artifacts.models import ArtifactCategory, RuntimeStatistics
from debusine.client.debusine import Debusine
from debusine.client.models import ArtifactResponse
from debusine.tasks.models import BackendType, BaseTaskDataModel

_backends: dict[BackendType, type["ExecutorInterface"]] = {}


class InstanceRunning(Exception):
    """Raised if the instance tries to start a started instance."""


class InstanceNotRunning(Exception):
    """Raised if the instance tries to use an instance that is not running."""


class ImageNotDownloadedError(Exception):
    """Raised if an Executor tries to use an image before downloading it."""


class ImageImportError(Exception):
    """Raised if an Executor fails to import an environment image."""


class ExecutorImageCategory(StrEnum):
    """Artifact categories that executors may expect for images."""

    TARBALL = str(ArtifactCategory.SYSTEM_TARBALL)
    IMAGE = str(ArtifactCategory.SYSTEM_IMAGE)


class ExecutorStatistics(BaseTaskDataModel):
    """
    Statistics for an executor.

    Similar to :py:class:`debusine.tasks.models.RuntimeStatistics`, but
    decoupled from that since tasks involve more than their executor.
    """

    #: Maximum disk space used, in bytes.
    disk_space: int | None = None

    #: Maximum RAM used, in bytes.
    memory: int | None = None

    #: Available disk space, in bytes; may be rounded.  (This is usually a
    #: property of the worker, but virtual machines can have more limited
    #: disk space.)
    available_disk_space: int | None = None

    #: Available RAM, in bytes; may be rounded.  (This is usually a property
    #: of the worker, but virtual machines can have more limited RAM.)
    available_memory: int | None = None

    #: Number of CPU cores.  (This is usually a property of the worker, but
    #: virtual machines can have different numbers of cores.)
    cpu_count: int | None = None

    def add_to(self, stats: RuntimeStatistics) -> None:
        """Update ``stats`` with our values."""
        # We're interested in the high-water mark of these statistics.
        if self.disk_space is not None:
            if stats.disk_space is None:
                stats.disk_space = self.disk_space
            else:
                stats.disk_space = max(stats.disk_space, self.disk_space)

        if self.memory is not None:
            if stats.memory is None:
                stats.memory = self.memory
            else:
                stats.memory = max(stats.memory, self.memory)

        # We record the initial values of these statistics, but do not
        # expect them to change afterwards.
        if (
            self.available_disk_space is not None
            and stats.available_disk_space is None
        ):
            stats.available_disk_space = self.available_disk_space

        if self.available_memory is not None and stats.available_memory is None:
            stats.available_memory = self.available_memory

        if self.cpu_count is not None and stats.cpu_count is None:
            stats.cpu_count = self.cpu_count


class ExecutorInterface(ABC):
    """
    Interface for managing an executor.

    All executors must support this minimal interface.
    """

    system_image: ArtifactResponse
    backend_name: BackendType
    image_category: ClassVar[ExecutorImageCategory]

    @override
    def __init_subclass__(
        cls, /, backend_name: BackendType, **kwargs: Any
    ) -> None:
        """Register the backend in _backends."""
        super().__init_subclass__(**kwargs)
        cls.backend_name = backend_name
        _backends[backend_name] = cls

    @abstractmethod
    def __init__(
        self,
        debusine_api: Debusine,
        system_image_id: int,
        workspace: str,
        working_directory: Path,
    ) -> None:
        """
        Instantiate an ExecutorInterface.

        :param debusine_api: The object to use the debusine client API
        :param system_image_id: An artifact ID pointing to the system
          tarball or disk image to use (as appropriate for each technology)
        """

    @classmethod
    @abstractmethod
    def available(cls) -> bool:
        """Determine whether this executor is available for operation."""

    @abstractmethod
    def download_image(self) -> str:
        """
        Make the image available locally.

        Fetch the image from artifact storage, if it isn't already available,
        and make it available locally.

        Return a path to the image or name, as appropriate for the backend.
        """

    @abstractmethod
    def image_name(self) -> str:
        """Return the path or name as returned from download_image()."""

    @abstractmethod
    def create(self) -> "InstanceInterface":
        """
        Create an InstanceInterface instance, using the image.

        Returns a new, stopped InstanceInterface.

        Each of these instances are pristine copies of the image, without any
        shared state between them.
        """

    @abstractmethod
    def autopkgtest_virt_server(self) -> str:
        """Return the name of the autopkgtest-virt-server for this backend."""

    @abstractmethod
    def autopkgtest_virt_args(self) -> list[str]:
        """
        Generate the arguments to drive an autopkgtest-virt-server.

        Tasks using tools that support the autopkgtest-virt-server interface
        can call this to generate the correct commandline for using this image.
        """

    @classmethod
    @abstractmethod
    def get_statistics(
        cls, instance: Optional["InstanceInterface"]
    ) -> ExecutorStatistics:
        """
        Return statistics for this executor.

        The worker may start collecting statistics as soon as the task
        starts, before an instance has been created.  Some statistics may
        not be available until an instance has been created, depending on
        the executor.
        """


class InstanceInterface(ABC):
    """
    Interface for execution within an instance of an execution environment.

    All execution environments support this minimal interface.

    Instances are ephemeral containers / VMs.
    """

    _uid_map: dict[str, int]
    _gid_map: dict[str, int]
    non_root_user: str = "_debusine"

    def __init__(self) -> None:
        """Initialize an InstanceInterface."""
        self._uid_map = {"root": 0}
        self._gid_map = {"root": 0}

    @abstractmethod
    def is_started(self) -> bool:
        """Determine if the instance is started."""

    def start(self) -> None:
        """
        Start the environment, making it ready to run commands.

        VM backends and containers that run init will boot the instance.
        """
        if self.is_started():
            raise InstanceRunning("Cannot start a started instance")
        self.do_start()

    @abstractmethod
    def do_start(self) -> None:
        """
        Start the environment, making it ready to run commands.

        VM backends and containers that run init will boot the instance.
        Called by start().
        """

    def stop(self) -> None:
        """Stop the environment, cleaning up afterwards."""
        if not self.is_started():
            raise InstanceNotRunning("Cannot stop a stopped instance")
        self.do_stop()

    @abstractmethod
    def do_stop(self) -> None:
        """Stop the environment, cleaning up afterwards."""

    def restart(self) -> None:
        """Restart the environment."""
        if not self.is_started():
            raise InstanceNotRunning("Cannot restart a stopped instance")
        self.do_restart()

    @abstractmethod
    def do_restart(self) -> None:
        """Restart the environment."""

    def file_push(
        self,
        source: Path,
        target: PurePath,
        user: str = "root",
        group: str = "root",
        mode: int = 0o644,
    ) -> None:
        """
        Copy a file into the environment.

        source is a regular file.
        target is the target file-name within an existing directory in the
        instance.

        Timestamps are not expected to be retained.
        """
        if not self.is_started():
            raise InstanceNotRunning("Cannot push files to a stopped instance")
        if not target.is_absolute():
            raise ValueError(f"target must be an absolute path (was: {target})")
        uid = self._get_uid(user)
        gid = self._get_gid(group)
        self.do_file_push(source, target, uid, gid, mode)

    @abstractmethod
    def do_file_push(
        self, source: Path, target: PurePath, uid: int, gid: int, mode: int
    ) -> None:
        """
        Copy a file into the environment.

        source is a regular file.
        target is the target file-name within an existing directory in the
        instance.

        Timestamps are not expected to be retained.
        """

    def file_pull(self, source: PurePath, target: Path) -> None:
        """
        Copy a file out of the environment.

        source is a regular file.
        target is the target file-name within an existing directory on the
        host.

        Timestamps are not expected to be retained.
        """
        if not self.is_started():
            raise InstanceNotRunning(
                "Cannot pull files from a stopped instance"
            )
        if not source.is_absolute():
            raise ValueError(f"source must be an absolute path (was: {source})")
        self.do_file_pull(source, target)

    @abstractmethod
    def do_file_pull(self, source: PurePath, target: Path) -> None:
        """
        Copy a file out of the environment.

        source is a regular file.
        target is the target file-name within an existing directory on the
        host.

        Only regular files are supported.
        Timestamps are not expected to be retained.
        """

    def directory_push(
        self,
        source: Path,
        target: PurePath,
        user: str = "root",
        group: str = "root",
    ) -> None:
        """
        Copy a directory (recursively) into the environment.

        source is a directory.
        target is an existing directory that source will be copied into.
        Files will belong to user and group.
        File modes will be retained.

        Only regular files and directories are supported.
        Timestamps are not expected to be retained.
        """
        if not self.is_started():
            raise InstanceNotRunning("Cannot push files to a stopped instance")
        if not target.is_absolute():
            raise ValueError(f"target must be an absolute path (was: {target})")
        uid = self._get_uid(user)
        gid = self._get_gid(group)
        self.do_directory_push(source, target, uid, gid)

    def do_directory_push(
        self, source: Path, target: PurePath, uid: int, gid: int
    ) -> None:
        """
        Copy a directory (recursively) into the environment.

        source is a directory.
        target is an existing directory that source will be copied into.
        Files will belong to user and group.
        File modes will be retained.

        Only regular files and directories are supported.
        Timestamps are not expected to be retained.
        """
        mode = S_IMODE(source.stat().st_mode)
        self.do_mkdir(target / source.name, uid, gid, mode, parents=False)

        for dir_path, dir_names, file_names in source.walk():
            target_relative_dir_path = dir_path.relative_to(source.parent)

            for dir_name in dir_names:
                source_dir_path = dir_path / dir_name
                target_dir = target / target_relative_dir_path / dir_name
                mode = S_IMODE(source_dir_path.stat().st_mode)
                self.do_mkdir(target_dir, uid, gid, mode, parents=False)

            for file_name in file_names:
                source_file_path = dir_path / file_name
                target_file_path = target / target_relative_dir_path / file_name
                mode = S_IMODE(source_file_path.stat().st_mode)

                self.do_file_push(
                    source_file_path, target_file_path, uid, gid, mode
                )

    def mkdir(
        self,
        target: PurePath,
        user: str = "root",
        group: str = "root",
        mode: int = 0o755,
        parents: bool = False,
    ) -> None:
        """
        Create directory inside the instance.

        Owned by user and group, with the specified mode.
        Create parent directories if necessary if parents is True.

        If target exists: raise FileExistsError().
        """
        if not self.is_started():
            raise InstanceNotRunning(
                "Cannot make directories in a stopped instance"
            )
        if not target.is_absolute():
            raise ValueError(f"target must be an absolute path (was: {target})")
        uid = self._get_uid(user)
        gid = self._get_gid(group)
        self.do_mkdir(target, uid, gid, mode, parents)

    def do_mkdir(
        self, target: PurePath, uid: int, gid: int, mode: int, parents: bool
    ) -> None:
        """
        Create directory inside the instance.

        Owned by user and group, with the specified mode.
        Create parent directories if necessary if parents is True.

        If target exists: raise FileExistsError().
        """
        cmd = ["mkdir", f"--mode={mode:o}"]
        if parents:
            cmd.append("--parents")
        cmd.append(str(target))
        try:
            self._run_check_output(cmd)
        except CalledProcessError as e:
            stderr = e.stderr or ''
            if e.returncode == 1 and stderr.endswith("File exists"):
                raise FileExistsError(
                    f'Cannot create directory: "{target}" already exists'
                )
            raise
        if uid != 0 and gid != 0:
            self._run_check_output(["chown", f"{uid}:{gid}", str(target)])

    @overload
    def run(
        self,
        args: list[str],
        *,
        text: Literal[True],
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[str]: ...

    @overload
    def run(
        self,
        args: list[str],
        *,
        text: Literal[False] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[bytes]: ...

    def run(
        self,
        args: list[str],
        **kwargs: Any,
    ) -> CompletedProcess[AnyStr]:
        """
        Run a command in the environment.

        Arguments behave as if passed to `subprocess.run`.
        """
        if not self.is_started():
            raise InstanceNotRunning(
                "Cannot run commands on a stopped instance"
            )
        kwargs.setdefault("stderr", PIPE)
        kwargs.setdefault("stdout", PIPE)
        return self.do_run(args, **kwargs)

    @overload
    @abstractmethod
    def do_run(
        self,
        args: list[str],
        text: Literal[True],
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[str]: ...

    @overload
    @abstractmethod
    def do_run(
        self,
        args: list[str],
        text: Literal[False] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[bytes]: ...

    @abstractmethod
    def do_run(
        self,
        args: list[str],
        text: Literal[False] | Literal[True] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[AnyStr]:
        """
        Run a command in the environment.

        Arguments behave as if passed to `subprocess.run`.
        """

    def _run_check_output(self, args: list[str]) -> str:
        """Run args as root in the instance, return output."""
        return self.run(
            args,
            text=True,
            run_as_root=True,
            stdout=PIPE,
            stderr=PIPE,
            check=True,
        ).stdout

    def _get_uid(self, username: str) -> int:
        """Look up the User ID for username inside the instance."""
        if username not in self._uid_map:
            try:
                entry = self._run_check_output(["getent", "passwd", username])
            except CalledProcessError as e:
                if e.returncode == 2:
                    raise KeyError(f"User not found in instance: {username}")
                raise
            uid = int(entry.split(":")[2])
            self._uid_map[username] = uid
        return self._uid_map[username]

    def _get_gid(self, group: str) -> int:
        """Look up the Group ID for group inside the instance."""
        if group not in self._gid_map:
            try:
                entry = self._run_check_output(["getent", "group", group])
            except CalledProcessError as e:
                if e.returncode == 2:
                    raise KeyError(f"Group not found in instance: {group}")
                raise
            gid = int(entry.split(":")[2])
            self._gid_map[group] = gid
        return self._gid_map[group]

    def create_user(self, username: str | None = None) -> int:
        """Provision a non-root user (if necessary) and return its user id."""
        if username is None:
            username = self.non_root_user
        try:
            return self._get_uid(username)
        except KeyError:
            pass

        self._run_check_output(
            [
                "sh",
                "-c",
                (
                    "if [ ! -x /usr/sbin/useradd ]; then "
                    "apt-get update && apt-get -y install passwd; "
                    "fi"
                ),
            ]
        )
        self._run_check_output(["useradd", username])
        return self._get_uid(username)


def executor_class(
    backend: BackendType,
) -> type[ExecutorInterface]:
    """
    Get the Executor class for a backend name.

    :param backend: The name of the executor backend.
    :return: An ExecutorInterface class.
    """
    if backend not in _backends:
        raise NotImplementedError(
            f"Support for backend {backend} is not yet implemented."
        )

    return _backends[backend]


def executor_backends() -> list[BackendType]:
    """List all registered executor backends."""
    return list(_backends.keys())


def analyze_worker_all_executors() -> dict[str, Any]:
    """
    Return dictionary with metadata for each executor backend.

    This method is called on the worker to collect information about the
    worker. The information is stored as a set of key-value pairs in a
    dictionary.

    That information is then reused on the scheduler to be fed to
    :py:meth:`debusine.tasks.BaseExternalTask.can_run_on` and determine if a
    task is suitable to be executed on the worker.
    """
    return {
        f"executor:{backend}:available": executor_class.available()
        for backend, executor_class in _backends.items()
    }


def executor_provided_tags() -> set[str]:
    """
    Return a set of scheduler tags provided by executors.

    This is used to signal executors' availability and feature sets.
    """
    # Prevent circular import
    import debusine.worker.tags as wtags

    return {
        wtags.EXECUTOR_PREFIX + backend
        for backend, executor_class in _backends.items()
        if executor_class.available()
    }



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for Debusine dput-ng utilities."""

from functools import partial
from textwrap import dedent
from unittest import mock

from debusine.client.config import ConfigHandler
from debusine.client.dput_ng.dput_ng_utils import make_debusine_client
from debusine.test import TestCase


class TestMakeDebusineClient(TestCase):
    """Test :py:func:`make_debusine_client`."""

    def test_properties(self) -> None:
        """The client has reasonable properties."""
        config_file_path = self.create_temporary_file()
        config_file_path.write_text(
            dedent(
                """\
                [server:debusine.example.net]
                api-url = https://debusine.example.net/api
                token = some-token
                scope = debian
                """
            )
        )
        profile = {
            "debusine_scope": "debian",
            "debusine_workspace": "base",
            "fqdn": "debusine.example.net",
        }

        with mock.patch(
            "debusine.client.dput_ng.dput_ng_utils.ConfigHandler",
            partial(ConfigHandler, config_file_path=config_file_path),
        ):
            client = make_debusine_client(profile)

        self.assertEqual(
            client.base_api_url, "https://debusine.example.net/api"
        )
        self.assertEqual(client.token, "some-token")
        self.assertEqual(client.scope, "debian")

    def test_searches_for_server(self) -> None:
        """The client searches for a server section with the right api-url."""
        config_file_path = self.create_temporary_file()
        config_file_path.write_text(
            dedent(
                """\
                [General]
                default-server = debusine.incus

                [server:debusine.incus]
                api-url = https://debusine.incus:8000/api
                token = some-token
                scope = debusine

                [server:example]
                api-url = https://debusine.example.net/api
                token = some-token
                scope = debian

                [server:another-example]
                api-url = https://debusine.another-example.net/api
                token = some-token
                scope = debian
                """
            )
        )
        profile = {
            "debusine_scope": "debian",
            "debusine_workspace": "base",
            "fqdn": "debusine.example.net",
        }

        with mock.patch(
            "debusine.client.dput_ng.dput_ng_utils.ConfigHandler",
            partial(ConfigHandler, config_file_path=config_file_path),
        ):
            client = make_debusine_client(profile)

        self.assertEqual(
            client.base_api_url, "https://debusine.example.net/api"
        )
        self.assertEqual(client.token, "some-token")
        self.assertEqual(client.scope, "debian")

    def test_missing_server(self) -> None:
        """The client reports an error if there is no suitable configuration."""
        config_file_path = self.create_temporary_file()
        config_file_path.write_text(
            dedent(
                """\
                [server:debusine.incus]
                api-url = https://debusine.incus:8000/api
                token = some-token
                scope = debusine
                """
            )
        )
        profile = {
            "debusine_scope": "debian",
            "debusine_workspace": "base",
            "fqdn": "debusine.example.net",
        }

        with (
            mock.patch(
                "debusine.client.dput_ng.dput_ng_utils.ConfigHandler",
                partial(ConfigHandler, config_file_path=config_file_path),
            ),
            self.assertRaisesRegex(
                ValueError,
                r"No Debusine client configuration for "
                r"'debusine\.example\.net/debian'; "
                r"run 'debusine setup' to configure it",
            ),
        ):
            make_debusine_client(profile)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Django-based command-line utility for administrative tasks."""

import os
import sys

from django.core.exceptions import ImproperlyConfigured


def main() -> None:
    """Run a management command."""
    basename = os.path.basename(sys.argv[0])

    # Drop privileges if running one of the packaged command wrappers as
    # root.
    match os.geteuid(), basename:
        case 0, "debusine-admin":
            os.execvp(
                "runuser",
                ["runuser", "--user=debusine-server", "--", *sys.argv],
            )
        case 0, "debusine-signing":
            os.execvp(
                "runuser",
                ["runuser", "--user=debusine-signing", "--", *sys.argv],
            )

    os.environ.setdefault(
        "DJANGO_SETTINGS_MODULE",
        (
            "debusine.signing.settings"
            if basename == "debusine-signing"
            else "debusine.project.settings"
        ),
    )

    # Must only be imported after DJANGO_SETTINGS_MODULE is set.
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:  # pragma: no cover
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc

    try:
        execute_from_command_line(sys.argv)
    except ImproperlyConfigured as exc:
        print("Improperly configured error:", exc, file=sys.stderr)
        sys.exit(3)
    except ValueError as exc:
        if isinstance(exc.__cause__, PermissionError):
            filename = exc.__cause__.filename or "Unknown"
            print(
                f"Permission error: {exc}. Check that the user running "
                f'{basename} has access to the file "{filename}".',
                file=sys.stderr,
            )
            sys.exit(3)
        else:
            raise



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the prometheus redis implementation."""

import time as tm
from threading import Event
from typing import override
from unittest import mock

from django.test import SimpleTestCase, override_settings
from fakeredis import FakeRedis
from redis import Redis

from debusine.server.prometheus_client.redis import (
    _daemon_threads,
    _keep_key_from_expiring,
    mark_process_dead,
    private_fake_redis,
    redis_client,
)


class RedisClientTestCase(SimpleTestCase):
    """Test redis_client()."""

    @override_settings(PROMETHEUS_REDIS_URL="redis://localhost/0")
    def test_redis(self) -> None:
        client = redis_client()
        self.assertIsInstance(client, Redis)
        self.assertNotIsInstance(client, FakeRedis)

    @private_fake_redis()
    def test_fake_redis(self) -> None:
        client = redis_client()
        self.assertIsInstance(client, FakeRedis)


@private_fake_redis()
class KeepMetricsAliveTestCase(SimpleTestCase):
    """Test KeepMetricsAliveThread and friends."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.stop_event = Event()

    @override
    def tearDown(self) -> None:
        super().tearDown()
        self.stop_event.set()

    def mock_loop(self) -> None:
        self.loop_event = Event()
        self.waiting_event = Event()
        patcher = mock.patch(
            (
                "debusine.server.prometheus_client.redis."
                "KeepMetricsAliveThread.loop_wait"
            ),
            side_effect=self.loop_wait,
        )
        self.addCleanup(patcher.stop)
        patcher.start()

    def loop_wait(self, _timeout: float) -> bool:
        self.waiting_event.set()
        self.waiting_event = Event()
        assert self.loop_event.wait(5) is True
        return self.stop_event.is_set()

    def run_loop_once(self, stop: bool = False) -> None:
        if stop:
            self.stop_event.set()
        self.loop_event.set()
        if not stop:
            self.loop_event = Event()
            assert self.waiting_event.wait() is True

    def test_mark_unknown_process_dead(self) -> None:
        mark_process_dead("unknown")

    def test_mark_known_process_dead(self) -> None:
        client = redis_client()
        client.set("key", "hello")
        _keep_key_from_expiring("identifier", "key")
        self.assertIn("identifier", _daemon_threads)
        mark_process_dead("identifier")
        self.assertNotIn("identifier", _daemon_threads)

    def test_live_daemon_updates_expiry(self) -> None:
        """Exercise the full lifecycle of the daemon thread."""
        self.mock_loop()
        client = redis_client()
        client.set("key", "hello")
        _keep_key_from_expiring("identifier", "key")
        self.assertLess(client.expiretime("key"), 0)

        unixtime = tm.time()
        self.run_loop_once()
        self.assertGreater(client.expiretime("key"), unixtime)
        self.run_loop_once(stop=True)



# Generated by Django 4.2.20 on 2025-05-25 19:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0007_alter_collectionitem_created_at_and_more'),
    ]
    replaces = [
        ('db', '0149_collectionitem_db_ci_name_idx'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='collectionitem',
            index=models.Index(models.F('name'), name='db_ci_name_idx'),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for signon views."""

import contextlib
import datetime as dt
import html
import re
from collections.abc import Generator, Sequence
from typing import Any, cast, override
from unittest import mock
from urllib.parse import parse_qs, urlparse

from django.contrib.auth.models import AnonymousUser
from django.contrib.messages import get_messages
from django.http import HttpRequest
from django.test import Client, RequestFactory, override_settings
from django.urls import resolve, reverse
from django.utils import timezone
from django_prometheus.testutils import assert_metric_diff, save_registry

from debusine.db.models import Identity, User
from debusine.db.playground import scenarios
from debusine.server.signon import providers
from debusine.server.signon.middleware import RequestSignonProtocol
from debusine.server.signon.providers import BoundOIDCProvider
from debusine.server.signon.signon import MapIdentityFailed, Signon
from debusine.server.signon.tests.test_signon import MockSession
from debusine.test.django import TestCase, TestResponseType


@override_settings(
    SIGNON_PROVIDERS=[
        providers.GitlabProvider(
            name="salsa",
            label="Salsa",
            url="https://salsa.debian.org",
            client_id="clientid",
            client_secret="clientsecret",
        ),
    ],
)
class SalsaSignonViews(TestCase):
    """Test Signon views."""

    scenario = scenarios.DefaultScopeUser()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.registry = save_registry()

    def _collect_urls(
        self, response: TestResponseType, match_re: str
    ) -> list[str]:
        """Collect external URLs in the response contents."""
        urls: list[str] = []
        re_url = re.compile(b'<a href="([^"]+)"')
        re_match = re.compile(match_re)
        for a in re_url.finditer(response.content):
            target = html.unescape(a.group(1).decode())
            if re_match.search(target):  # pragma: no cover
                # Excluding from coverage because this is currently always
                # true, but the check is here to be future-proof in case the
                # login page adds further links in other parts of the templates
                urls.append(target)

        return urls

    def _make_identity(
        self,
        user: User | None = None,
        issuer: str = "salsa",
        subject: str | None = None,
        **kwargs: Any,
    ) -> Identity:
        """Create a test identity."""
        if subject is None:
            subject = f"{issuer}@debian.org"
        return Identity.objects.create(
            user=user,
            issuer=issuer,
            subject=subject,
            claims={},
            **kwargs,
        )

    def _make_post_request(self, url: str) -> HttpRequest:
        """Create a request that can be manipulated before invoking a view."""
        factory = RequestFactory()
        request = factory.post(url)
        request.user = AnonymousUser()
        request.session = MockSession()
        cast(RequestSignonProtocol, request).signon = Signon(request)
        # Bypass CSRF checks for logout.  django.test.Client does this, but
        # we're bypassing that in order to manipulate the request, so for
        # now we have to set a private attribute.
        setattr(request, "_dont_enforce_csrf_checks", True)
        return request

    def test_login_links(self) -> None:
        """Login form shows link to authenticate with salsa."""
        response = self.client.get(reverse("login"))

        # Collect urls to the salsa provider
        external_urls = self._collect_urls(
            response, r"salsa\.debian\.org/oauth"
        )
        self.assertEqual(len(external_urls), 1)

        url = urlparse(external_urls[0])

        # Test the components of the authorize URL
        self.assertEqual(url.scheme, "https")
        self.assertEqual(url.netloc, "salsa.debian.org")
        self.assertEqual(url.path, "/oauth/authorize")

        query = parse_qs(url.query)
        self.assertEqual(query["response_type"], ["code"])
        self.assertEqual(
            query["redirect_uri"],
            [
                "http://testserver"
                + reverse("signon:oidc_callback", kwargs={"name": "salsa"})
            ],
        )
        self.assertIn("client_id", query)
        self.assertIn("state", query)

    def test_login_links_when_signed_in(self) -> None:
        """Active identities don't show a login link."""
        identity = self._make_identity(user=self.scenario.user)

        # Simulate a session with a logged in user
        url = reverse("login")
        request = self._make_post_request(url)
        assert isinstance(request, RequestSignonProtocol)
        request.signon.activate_identity(identity)
        self.assertEqual(request.user, self.scenario.user)
        self.scenario.set_current()

        view = resolve(url).func
        response = view(request)
        response.render()

        # Collect urls to the salsa provider
        external_urls = self._collect_urls(
            response, r"salsa\.debian\.org/oauth"
        )
        self.assertEqual(len(external_urls), 0)

    def test_login_without_provider_icon(self) -> None:
        response = self.client.get(reverse("login"))

        tree = self.assertResponseHTML(response)
        login_item = self.assertHasElement(
            tree, "//li[@class='list-group-item list-group-item-action']"
        )
        self.assertFalse(login_item.xpath("img"))

    @override_settings(
        SIGNON_PROVIDERS=[
            providers.GitlabProvider(
                name="salsa",
                label="Salsa",
                icon="signon/salsa.svg",
                url="https://salsa.debian.org",
                client_id="clientid",
                client_secret="clientsecret",
            ),
        ],
    )
    def test_login_with_provider_icon(self) -> None:
        response = self.client.get(reverse("login"))

        tree = self.assertResponseHTML(response)
        icon = self.assertHasElement(
            tree, "//li[@class='list-group-item list-group-item-action']/img"
        )
        self.assertEqual(icon.get("alt"), "Salsa icon")
        self.assertEqual(icon.get("src"), "/static/signon/salsa.svg")

    def test_logout(self) -> None:
        """Logout also deactivates signon identities."""
        identity = self._make_identity(user=self.scenario.user)

        # Simulate a session with a logged in user
        url = reverse("logout")
        request = self._make_post_request(url)
        assert isinstance(request, RequestSignonProtocol)
        request.signon.activate_identity(identity)
        self.assertEqual(request.user, self.scenario.user)
        self.scenario.set_current()

        view = resolve(url).func
        response = view(request)
        response.render()

        # Log out happened
        self.assertContains(response, "You have been logged out")

        # User is not authenticated anymore
        self.assertFalse(request.user.is_authenticated)

        # Identities have been deactivated
        self.assertIsNone(request.session.get("signon_identity_salsa"))
        assert isinstance(request, RequestSignonProtocol)
        self.assertEqual(request.signon.identities, {})

    @contextlib.contextmanager
    def setup_mock_auth(
        self,
        identity: Identity | None = None,
        options: Sequence[str] = (),
        **claims: Any,
    ) -> Generator[None]:
        """
        Mock remote authentication.

        :param identity: if provided it is used to fill default claims
        """
        if identity is not None:
            claims.setdefault("sub", identity.subject)

        def load_tokens(self: BoundOIDCProvider) -> None:
            self.id_token_claims = claims
            self.options = options

        with mock.patch(
            "debusine.server.signon.providers.BoundOIDCProvider.load_tokens",
            side_effect=load_tokens,
            autospec=True,
        ):
            yield

    def test_oidc_callback_create_identity(self) -> None:
        """The OIDC callback creates a missing identity."""
        self.assertEqual(Identity.objects.count(), 0)

        cb_url = reverse("signon:oidc_callback", kwargs={"name": "salsa"})

        start_time = timezone.now()

        with self.setup_mock_auth(sub="123", profile="profile_url"):
            response = self.client.get(cb_url)

        self.assertRedirects(response, reverse("homepage:homepage"))

        # Check that the Identity has been created and fully populated
        self.assertEqual(Identity.objects.count(), 1)

        identities = list(Identity.objects.all())
        self.assertIsNone(identities[0].user)
        self.assertEqual(identities[0].issuer, "salsa")
        self.assertEqual(identities[0].subject, "123")
        self.assertGreaterEqual(identities[0].last_used, start_time)
        self.assertEqual(
            identities[0].claims, {"sub": "123", "profile": "profile_url"}
        )

        # Check that the Identity is attached to the session
        self.assertEqual(
            self.client.session["signon_identity_salsa"], identities[0].pk
        )

        # The identity is activated
        assert isinstance(response.wsgi_request, RequestSignonProtocol)
        self.assertEqual(
            response.wsgi_request.signon.identities, {"salsa": identities[0]}
        )
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer="salsa",
            activated="False",
        )

        # The user is not authenticated, because the identity is not bound
        self.assertFalse(response.wsgi_request.user.is_authenticated)

    def test_oidc_callback_login(self) -> None:
        """The OIDC callback logs in an existing bound identity."""
        self.assertEqual(Identity.objects.count(), 0)

        ident = self._make_identity(
            user=self.scenario.user,
            subject="123",
            last_used=timezone.now() - dt.timedelta(days=1),
        )

        cb_url = reverse("signon:oidc_callback", kwargs={"name": ident.issuer})

        start_time = timezone.now()
        with self.setup_mock_auth(identity=ident, profile="profile_url"):
            response = self.client.get(cb_url)

        self.assertRedirects(response, reverse("homepage:homepage"))

        # Check that no new Identity has been created
        self.assertEqual(Identity.objects.count(), 1)

        ident.refresh_from_db()
        # last_used has been updated
        self.assertGreaterEqual(ident.last_used, start_time)
        # Claims in ident have been filled
        self.assertEqual(ident.claims, {"profile": "profile_url", "sub": "123"})

        # Check that the Identity is attached to the session
        self.assertEqual(self.client.session["signon_identity_salsa"], ident.pk)

        # The identity is activated
        assert isinstance(response.wsgi_request, RequestSignonProtocol)
        self.assertEqual(
            response.wsgi_request.signon.identities, {"salsa": ident}
        )
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer="salsa",
            activated="True",
        )

        # The user has been correctly authenticated
        self.assertTrue(response.wsgi_request.user.is_authenticated)
        self.assertEqual(response.wsgi_request.user, self.scenario.user)

    def test_oidc_callback_login_twice(self) -> None:
        self.assertEqual(Identity.objects.count(), 0)

        last_used = timezone.now() - dt.timedelta(days=1)
        ident = self._make_identity(
            user=self.scenario.user,
            subject="123",
            last_used=last_used,
        )

        cb_url = reverse("signon:oidc_callback", kwargs={"name": ident.issuer})

        self.client.force_login(self.scenario.user)

        start_time = timezone.now()
        with self.setup_mock_auth(identity=ident, profile="profile_url"):
            response = self.client.get(cb_url)

        self.assertEqual(response.status_code, 403)

        # Check that no new Identity has been created
        self.assertEqual(Identity.objects.count(), 1)

        ident.refresh_from_db()
        # last_used has been updated
        self.assertGreaterEqual(ident.last_used, start_time)
        # Claims in ident have been filled
        self.assertEqual(ident.claims, {"profile": "profile_url", "sub": "123"})

        # Check that the Identity is attached to the session
        self.assertNotIn("signon_identity_salsa", self.client.session)

        # The identity is not activated
        assert isinstance(response.wsgi_request, RequestSignonProtocol)
        self.assertEqual(response.wsgi_request.signon.identities, {})
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer="salsa",
            activated="False",
        )

    def test_oidc_callback_login_user_inactive(self) -> None:
        """Do not log in an inactive user."""
        self.assertEqual(Identity.objects.count(), 0)

        self.scenario.user.is_active = False
        self.scenario.user.save()
        last_used = timezone.now() - dt.timedelta(days=1)
        ident = self._make_identity(
            user=self.scenario.user,
            subject="123",
            last_used=last_used,
        )

        cb_url = reverse("signon:oidc_callback", kwargs={"name": ident.issuer})

        start_time = timezone.now()
        with self.setup_mock_auth(identity=ident, profile="profile_url"):
            response = self.client.get(cb_url)

        self.assertEqual(response.status_code, 403)

        # Check that no new Identity has been created
        self.assertEqual(Identity.objects.count(), 1)

        ident.refresh_from_db()
        # last_used has been updated
        self.assertGreaterEqual(ident.last_used, start_time)
        # Claims in ident have been filled
        self.assertEqual(ident.claims, {"profile": "profile_url", "sub": "123"})

        # Check that the Identity is attached to the session
        self.assertNotIn("signon_identity_salsa", self.client.session)

        # The identity is not activated
        assert isinstance(response.wsgi_request, RequestSignonProtocol)
        self.assertEqual(response.wsgi_request.signon.identities, {})
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer="salsa",
            activated="False",
        )

        # The user is still marked as inactive
        self.scenario.user.refresh_from_db()
        self.assertFalse(self.scenario.user.is_active)

    def test_oidc_callback_bind(self) -> None:
        """The OIDC callback performs bind if requested."""
        self.assertEqual(Identity.objects.count(), 0)

        ident = self._make_identity(
            subject="123",
            last_used=timezone.now() - dt.timedelta(days=1),
        )

        cb_url = reverse("signon:oidc_callback", kwargs={"name": ident.issuer})

        self.client.force_login(self.scenario.user)

        start_time = timezone.now()
        with self.setup_mock_auth(
            identity=ident, options=["bind"], profile="profile_url"
        ):
            response = self.client.get(cb_url)

        self.assertRedirects(response, reverse("homepage:homepage"))

        # Check that no new Identity has been created
        self.assertEqual(Identity.objects.count(), 1)

        ident.refresh_from_db()
        # last_used has been updated
        self.assertGreaterEqual(ident.last_used, start_time)
        # Claims in ident have been filled
        self.assertEqual(ident.claims, {"profile": "profile_url", "sub": "123"})
        # Identity has been bound
        self.assertEqual(ident.user, self.scenario.user)
        # User is still logged in
        self.assertEqual(response.wsgi_request.user, self.scenario.user)

        # Check that the Identity is attached to the session
        assert isinstance(response.wsgi_request, RequestSignonProtocol)
        self.assertEqual(
            response.wsgi_request.signon.identities, {"salsa": ident}
        )
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer="salsa",
            activated="True",
        )

    def test_oidc_callback_wrong_provider(self) -> None:
        """The OIDC callback with a wrong provider."""
        self.assertEqual(Identity.objects.count(), 0)

        cb_url = reverse("signon:oidc_callback", kwargs={"name": "wrong"})

        response = self.client.get(cb_url)
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Identity.objects.count(), 0)
        self.assertFalse(response.wsgi_request.user.is_authenticated)
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer="wrong",
            activated="False",
        )

    def test_oidc_callback_validation_error(self) -> None:
        ident = self._make_identity(
            subject="123",
            last_used=timezone.now() - dt.timedelta(days=1),
        )
        cb_url = reverse("signon:oidc_callback", kwargs={"name": ident.issuer})

        with self.assertLogs("debusine.server.signon") as log:
            response = self.client.get(cb_url)
        self.assertEqual(response.status_code, 403)

        self.assertFalse(response.wsgi_request.user.is_authenticated)
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer=ident.issuer,
            activated="False",
        )
        self.assertEqual(
            log.output,
            [
                f"INFO:debusine.server.signon:OIDC validation for provider "
                f"'{ident.issuer}' failed: "
                f"Request state mismatch: expected state not found in session"
            ],
        )

    def test_oidc_callback_map_identity_to_user_failed(self) -> None:
        """Test user interaction when activating an identity fails."""
        ident = self._make_identity(
            subject="123",
            last_used=timezone.now() - dt.timedelta(days=1),
        )
        cb_url = reverse("signon:oidc_callback", kwargs={"name": ident.issuer})
        client = Client()
        with (
            self.setup_mock_auth(identity=ident, profile="profile_url"),
            mock.patch(
                "debusine.server.signon.signon.Signon.identity_create_user",
                side_effect=MapIdentityFailed("test message"),
            ),
        ):
            response = client.get(cb_url)
        self.assertRedirects(response, reverse("homepage:homepage"))
        messages = [m.message for m in get_messages(response.wsgi_request)]
        self.assertEqual(messages, ["Login failed: test message"])
        assert_metric_diff(
            self.registry,
            1,
            "debusine_signon_oidc_callbacks_total",
            issuer="salsa",
            activated="False",
        )

        ident.refresh_from_db()
        self.assertIsNone(ident.user)

    def test_oidc_callback_redirect_url(self) -> None:
        """Test redirect after successful OIDC callback."""
        ident = self._make_identity(
            user=self.scenario.user,
            subject="123",
            last_used=timezone.now() - dt.timedelta(days=1),
        )

        cb_url = reverse("signon:oidc_callback", kwargs={"name": ident.issuer})

        # Default redirect
        client = Client()
        with self.setup_mock_auth(identity=ident, profile="profile_url"):
            response = client.get(cb_url)
        self.assertRedirects(response, reverse("homepage:homepage"))

        # Redirect comes from SIGNON_DEFAULT_REDIRECT
        with override_settings(SIGNON_DEFAULT_REDIRECT="https://example.org"):
            client = Client()
            with self.setup_mock_auth(identity=ident, profile="profile_url"):
                response = client.get(cb_url)
            self.assertRedirects(
                response, "https://example.org", fetch_redirect_response=False
            )

        # Redirect from the session
        client = Client()
        session = client.session
        session["sso_callback_next_url"] = "https://session.example.org"
        session.save()
        with self.setup_mock_auth(identity=ident, profile="profile_url"):
            response = client.get(cb_url)
        self.assertRedirects(
            response,
            "https://session.example.org",
            fetch_redirect_response=False,
        )
        self.assertNotIn("sso_callback_next_url", client.session)

    def test_bindidentity(self) -> None:
        """Bind identity redirects correctly."""
        self.client.force_login(self.scenario.user)
        session = self.client.session
        response = self.client.get(
            reverse("signon:bind_identity", kwargs={"name": "salsa"})
        )
        self.assertEqual(response.status_code, 302)
        url = response.headers["Location"]
        parsed = urlparse(url)
        self.assertEqual(parsed.scheme, "https")
        self.assertEqual(parsed.netloc, "salsa.debian.org")
        self.assertEqual(parsed.path, "/oauth/authorize")
        args = parse_qs(parsed.query)
        state = args["state"][0]
        self.assertEqual(session["signon_state_salsa"], state)
        self.assertEqual(session["signon_state_salsa_options"], ["bind"])

    def test_bindidentity_no_user(self) -> None:
        """Bind identity fails if not logged in."""
        response = self.client.get(
            reverse("signon:bind_identity", kwargs={"name": "salsa"})
        )
        self.assertEqual(response.status_code, 403)

    def test_bindidentity_wrong_provider(self) -> None:
        """Bind identity fails for a nonexisting provider."""
        self.client.force_login(self.scenario.user)
        response = self.client.get(
            reverse("signon:bind_identity", kwargs={"name": "wrong"})
        )
        self.assertEqual(response.status_code, 403)

    # TODO: when we will support more than one authentication provider, we can
    # implement a profile view to link another external provider to the current
    # user.
    #
    # For now we only support a single authentication provider, and we have no
    # urgent need of this

    # def test_bind_view(self) -> None:
    #     """Test user profile view to bind external providers."""
    #     raise NotImplementedError()

    # def test_oidc_callback_bind_intent(self) -> None:
    #     """Test callback when authenticated with bind intent."""
    #     raise NotImplementedError()

    # def test_oidc_callback_no_bind_intent(self) -> None:
    #     """Test callback when authenticated with no bind intent."""
    #     raise NotImplementedError()



# Generated by Django 4.2.28 on 2026-03-24 16:29

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0046_copy_build_arch_to_tag'),
    ]

    operations = [
        migrations.AlterField(
            model_name='worker',
            name='activation_token',
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='activating_worker',
                to='db.token',
            ),
        ),
    ]
